from amuse.units import units
from amuse.datamodel import new_cartesian_grid
from amuse.support.exceptions import AmuseException
from amuse.test.amusetest import TestCase

try:
    from omuse.ext import grid_remappers
except:
    from amuse.ext import grid_remappers

import numpy

from amuse.datamodel.grids import *
from amuse.datamodel.staggeredgrid import StaggeredGrid


class TestGridRemappers(TestCase):

    def setUp(self):
        if not grid_remappers.matplotlib_available:
            self.skip("matplotlib not available")

    def test1(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((5, 10), 2., offset=[0., 0.15])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.interpolating_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)

    def test2(self):
        source = new_cartesian_grid((10, 20), 1. | units.m)
        target = new_cartesian_grid((5, 10), 2. | units.m, offset=[0., 0.15] | units.m)

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.interpolating_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)


class TestGridRemappers_bilinear(TestCase):

    def test1(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((5, 10), 2., offset=[0., 0.25])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.bilinear_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)

    def test2(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((8, 8), 0.125, offset=[0.5, 0.5])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.bilinear_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)

    def test3(self):
        source = new_cartesian_grid((10, 20), 1. | units.m)
        target = new_cartesian_grid((5, 10), 2. | units.m, offset=[0., 0.25] | units.m)

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.bilinear_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)

    def test4(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((10, 20), 1.5, offset=[-0.5, -0.5])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.bilinear_2D_remapper(source, target, check_inside=False)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.xcopy, numpy.clip(target.x, 0.5, 9.5))
        self.assertEqual(target.ycopy, numpy.clip(target.y, 0.5, 19.5))

    def test5(self):
        source = new_regular_grid((10, 20, 5), [1.*10, 1.*20, 1.])
        target = new_regular_grid((10, 20, 5), [1.5*10, 1.5*20, 1.], offset=[-0.5, -0.5, 0.])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.bilinear_2D_remapper(source, target, check_inside=False, do_slices=True)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.xcopy, numpy.clip(target.x, 0.5, 9.5))
        self.assertEqual(target.ycopy, numpy.clip(target.y, 0.5, 19.5))


class TestGridRemappers_nearest(TestCase):

    def test1(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((5, 10), 1., offset=[0., 0.])

        source.xcopy = source.x
        source.ycopy = source.y

        remapper = grid_remappers.nearest_2D_remapper(source, target)
        remapper.forward_mapping(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)


class TestConservativeSphericalRemapper(TestCase):

    def setUp(self):
        try:
            from omuse.community.cdo.interface import CDORemapper
        except:
            self.skip("conservative spherical remapper requires omuse.community.cdo.interface")

        # this test creates a structured staggered grid and an unstructured staggered grid
        # and then uses the conservative_spherical_remapper to remap values between the grids

        # define nodal points and triangles of a small test grid
        # got this grid from http://matplotlib.org/examples/pylab_examples/triplot_demo.html
        xy = numpy.asarray([
            [-0.101, 0.872], [-0.080, 0.883], [-0.069, 0.888], [-0.054, 0.890],
            [-0.045, 0.897], [-0.057, 0.895], [-0.073, 0.900], [-0.087, 0.898],
            [-0.090, 0.904], [-0.069, 0.907], [-0.069, 0.921], [-0.080, 0.919],
            [-0.073, 0.928], [-0.052, 0.930], [-0.048, 0.942], [-0.062, 0.949],
            [-0.054, 0.958], [-0.069, 0.954], [-0.087, 0.952], [-0.087, 0.959],
            [-0.080, 0.966], [-0.085, 0.973], [-0.087, 0.965], [-0.097, 0.965],
            [-0.097, 0.975], [-0.092, 0.984], [-0.101, 0.980], [-0.108, 0.980],
            [-0.104, 0.987], [-0.102, 0.993], [-0.115, 1.001], [-0.099, 0.996],
            [-0.101, 1.007], [-0.090, 1.010], [-0.087, 1.021], [-0.069, 1.021],
            [-0.052, 1.022], [-0.052, 1.017], [-0.069, 1.010], [-0.064, 1.005],
            [-0.048, 1.005], [-0.031, 1.005], [-0.031, 0.996], [-0.040, 0.987],
            [-0.045, 0.980], [-0.052, 0.975], [-0.040, 0.973], [-0.026, 0.968],
            [-0.020, 0.954], [-0.006, 0.947], [0.003, 0.935], [0.006, 0.926],
            [0.005, 0.921], [0.022, 0.923], [0.033, 0.912], [0.029, 0.905],
            [0.017, 0.900], [0.012, 0.895], [0.027, 0.893], [0.019, 0.886],
            [0.001, 0.883], [-0.012, 0.884], [-0.029, 0.883], [-0.038, 0.879],
            [-0.057, 0.881], [-0.062, 0.876], [-0.078, 0.876], [-0.087, 0.872],
            [-0.030, 0.907], [-0.007, 0.905], [-0.057, 0.916], [-0.025, 0.933],
            [-0.077, 0.990], [-0.059, 0.993]])
        triangles = numpy.asarray([
            [67, 66,  1], [65,  2, 66], [1, 66,  2], [64,  2, 65], [63,  3, 64],
            [60, 59, 57], [2, 64,  3], [3, 63,  4], [0, 67,  1], [62,  4, 63],
            [57, 59, 56], [59, 58, 56], [61, 60, 69], [57, 69, 60], [4, 62, 68],
            [6,  5,  9], [61, 68, 62], [69, 68, 61], [9,  5, 70], [6,  8,  7],
            [4, 70,  5], [8,  6,  9], [56, 69, 57], [69, 56, 52], [70, 10,  9],
            [54, 53, 55], [56, 55, 53], [68, 70,  4], [52, 56, 53], [11, 10, 12],
            [69, 71, 68], [68, 13, 70], [10, 70, 13], [51, 50, 52], [13, 68, 71],
            [52, 71, 69], [12, 10, 13], [71, 52, 50], [71, 14, 13], [50, 49, 71],
            [49, 48, 71], [14, 16, 15], [14, 71, 48], [17, 19, 18], [17, 20, 19],
            [48, 16, 14], [48, 47, 16], [47, 46, 16], [16, 46, 45], [23, 22, 24],
            [21, 24, 22], [17, 16, 45], [20, 17, 45], [21, 25, 24], [27, 26, 28],
            [20, 72, 21], [25, 21, 72], [45, 72, 20], [25, 28, 26], [44, 73, 45],
            [72, 45, 73], [28, 25, 29], [29, 25, 31], [43, 73, 44], [73, 43, 40],
            [72, 73, 39], [72, 31, 25], [42, 40, 43], [31, 30, 29], [39, 73, 40],
            [42, 41, 40], [72, 33, 31], [32, 31, 33], [39, 38, 72], [33, 72, 38],
            [33, 38, 34], [37, 35, 38], [34, 38, 35], [35, 37, 36]])

        num_elems = len(triangles)
        elements = UnstructuredGrid(num_elems)
        elements.n1 = triangles[:, 0] - 1
        elements.n2 = triangles[:, 1] - 1
        elements.n3 = triangles[:, 2] - 1

        lons = numpy.zeros(num_elems, dtype=numpy.double)
        lats = numpy.zeros(num_elems, dtype=numpy.double)
        for i in range(num_elems):
            for n in triangles[i]:
                lons[i] += xy[n, 0]/3.0
                lats[i] += xy[n, 1]/3.0
        elements.lon = (lons | units.rad)
        elements.lat = (lats | units.rad)

        nodes = UnstructuredGrid(len(xy))
        nodes.lon = (xy[:, 0] | units.rad)
        nodes.lat = (xy[:, 1] | units.rad)

        self.unstructured = StaggeredGrid(elements, nodes)

        # generate corners for a simple structured grid as source grid
        shape = [5, 5]
        lon_range = xy[:, 0].max() - xy[:, 0].min() + 0.025
        lon_min = xy[:, 0].min() - 0.0125
        lat_range = xy[:, 1].max() - xy[:, 1].min() + 0.025
        lat_min = xy[:, 1].min() - 0.0125
        ind = numpy.indices((shape[0]+1, shape[1]+1))
        lats = numpy.array(ind[1], dtype=numpy.float)
        lats = lat_min + lats/shape[1] * lat_range
        lons = numpy.array(ind[0], dtype=numpy.float)
        lons = lon_min + lons/shape[0] * lon_range

        corners = numpy.array([lons, lats])
        elements = new_structured_grid(shape, corners, axes_names=['lon', 'lat'])
        nodes = StructuredGrid(*ind[0].shape)
        nodes.lat = (lats | units.rad)
        nodes.lon = (lons | units.rad)
        self.structured = StaggeredGrid(elements, nodes)

        corners += 0.01  # shift the grid
        nodes = StructuredGrid(*ind[0].shape)
        nodes.lat = (corners[1] | units.rad)
        nodes.lon = (corners[0] | units.rad)
        self.structured2 = StaggeredGrid(elements, nodes)

    def test1(self):
        target = self.unstructured
        source = self.structured

        # set some values on the grid
        constant_field = numpy.zeros(source.nodes.shape, dtype=numpy.double)
        constant_field += 1.0
        source.nodes.const = constant_field

        constant_field = numpy.zeros(source.elements.shape, dtype=numpy.double)
        constant_field += 1.0
        source.elements.const2 = constant_field

        # create remapper
        remapper = grid_remappers.conservative_spherical_remapper(source, target)

        # remap values
        remapper.forward_mapping(["const", "const2"])

        self.assertTrue(numpy.all(target.nodes.const >= 0.0), msg="Expecting all remapped values values to be larger than zero")
        self.assertTrue(numpy.all(target.elements.const2 >= 0.0), msg="Expecting all remapped values values to be larger than zero")

    def test2(self):
        target = self.structured2
        source = self.structured

        # set some values on the grid
        constant_field = numpy.zeros(source.nodes.shape, dtype=numpy.double)
        constant_field += 1.0
        source.nodes.const = constant_field

        constant_field = numpy.zeros(source.elements.shape, dtype=numpy.double)
        constant_field += 1.0
        source.elements.const2 = constant_field

        # create remapper
        remapper = grid_remappers.conservative_spherical_remapper(source, target)

        # remap values
        remapper.forward_mapping(["const", "const2"])

        self.assertTrue(numpy.all(target.nodes.const >= 0.0), msg="Expecting all remapped values values to be larger than zero")
        self.assertTrue(numpy.all(target.elements.const2 >= 0.0), msg="Expecting all remapped values values to be larger than zero")

    def test3(self):
        target = self.unstructured
        source = self.structured

        # set some values on the grid
        constant_field = numpy.zeros(source.nodes.shape, dtype=numpy.double)
        constant_field += 1.0
        source.nodes.const = constant_field

        constant_field = numpy.zeros(source.elements.shape, dtype=numpy.double)
        constant_field += 1.0
        source.elements.const2 = constant_field

        # use a remapping channel
        channel = source.new_remapping_channel_to(target, remapper=grid_remappers.conservative_spherical_remapper)

        # remap values
        channel.copy_attributes(["const", "const2"])

        self.assertTrue(numpy.all(target.nodes.const >= 0.0), msg="Expecting all remapped values values to be larger than zero")
        self.assertTrue(numpy.all(target.elements.const2 >= 0.0), msg="Expecting all remapped values values to be larger than zero")


class TestGridRemappingChannel(TestCase):
    def setUp(self):
        if not grid_remappers.matplotlib_available:
            self.skip("matplotlib not available")

    def test1(self):
        source = new_cartesian_grid((10, 20), 1.)
        target = new_cartesian_grid((5, 10), 2., offset=[0., 0.5])

        channel = source.new_remapping_channel_to(target, grid_remappers.interpolating_2D_remapper)

        source.xcopy = source.x
        source.ycopy = source.y

        channel.copy_attributes(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)

    def test2(self):
        source = new_cartesian_grid((10, 20), 1. | units.m)
        target = new_cartesian_grid((5, 10), 2. | units.m, offset=[0., 0.5] | units.m)

        channel = source.new_remapping_channel_to(target, grid_remappers.interpolating_2D_remapper)

        source.xcopy = source.x
        source.ycopy = source.y

        channel.copy_attributes(["xcopy", "ycopy"])
        self.assertEqual(target.x, target.xcopy)
        self.assertEqual(target.y, target.ycopy)



__revision__ = "$Id:$"

import sys, os, re, subprocess,stat

from stat import ST_MODE
from distutils import sysconfig
from distutils.core import Command
from distutils.dep_util import newer
from distutils.util import convert_path
from distutils import log

class generate_main(Command):

    description = "generate shell script to run amuse"

    user_options = [
        ('amuse-dir=', 'd', "root directory of the amuse project"),
    ]

    def initialize_options (self):
        self.amuse_dir = None

    def finalize_options (self):
        if self.amuse_dir is None:
            self.amuse_dir =os.path.dirname(os.path.dirname(__file__))

    def get_source_files(self):
        return self.latex_documents

    def run (self):
        test_directory = os.path.join(self.amuse_dir, 'test')
        src_directory = os.path.join(self.amuse_dir, 'src')

        with open('amuse.sh','w') as script_file:
            script_file.write('#!/bin/sh')
            script_file.write('\n\n')
            script_file.write('export PYTHONPATH=${PYTHONPATH}')
            for x in [test_directory, src_directory]:
                script_file.write(':')
                script_file.write(x)

            script_file.write('\n')
            script_file.write('export AMUSE_DIR=')
            script_file.write(self.amuse_dir)
            script_file.write('\n')
            script_file.write(sys.executable)
            script_file.write(' "$@"\n')
        os.chmod('amuse.sh', stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

        with open('iamuse.sh','w') as script_file:
            script_file.write('#!/usr/bin/env python')
            script_file.write('\n\n')
            script_file.write('import IPython.Shell\n')
            script_file.write('import sys\n')
            for x in [test_directory, src_directory]:
                script_file.write("sys.path.append('{0}')\n".format(x))
            script_file.write('amuse_root_dir = "')
            script_file.write(self.amuse_dir)
            script_file.write('"\n')
            script_file.write('IPython.Shell.start().mainloop()\n')
        os.chmod('iamuse.sh', stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)




import os
from amuse.test import amusetest

from amuse.datamodel import new_cartesian_grid, Particles

from amuse.io import read_set_from_file, write_set_to_file


class test_github856(amusetest.TestCase):
    def test1(self):
        filename = os.path.join(self.get_path_to_results(), "github856.amuse")

        g1 = new_cartesian_grid((5, 5), 1)
        write_set_to_file(g1, filename, "amuse")
        del g1
        g2 = read_set_from_file(filename, "amuse")
        self.assertEquals(g2.get_axes_names(), "xy")

    def test2(self):
        g1 = Particles(lon=[1, 2], lat=[3, 4])
        g1.add_vector_attribute("lonlat", ["lon", "lat"])
        g2 = g1.copy()
        self.assertEquals(g2.lonlat, [[1, 3], [2, 4]])

    def test3(self):
        filename = os.path.join(self.get_path_to_results(), "github856_2.amuse")

        g1 = Particles(lon=[1, 2], lat=[3, 4])
        g1.add_vector_attribute("lonlat", ["lon", "lat"])
        write_set_to_file(g1, filename, "amuse")
        del g1
        g2 = read_set_from_file(filename, "amuse")
        self.assertEquals(g2.lonlat, [[1, 3], [2, 4]])



import numpy
from numpy import random

from amuse.test.amusetest import get_path_to_results, TestCase
from amuse.units import units, nbody_system
from amuse.support.exceptions import AmuseException
from amuse.datamodel.particles import Particles
from amuse.ic.plummer import new_plummer_model
from amuse.community.hop.interface import HopInterface, Hop


class TestHopInterface(TestCase):
    def test1(self):
        print("First test: adding particles, setting and getting.")
        hop = HopInterface()
        hop.initialize_code()
        n, err = hop.get_number_of_particles()
        self.assertEqual(n, 0)
        self.assertEqual(err, 0)

        for i in range(6):
            id, err = hop.new_particle(0, i*i, 0, 0)
            n, err = hop.get_number_of_particles()
            self.assertEqual(n, i+1)
            self.assertEqual(err, 0)

        for i in range(6):
            x, y, z, err = hop.get_position(i)
            self.assertEqual(x, i*i)
            self.assertEqual(y, 0)
            self.assertEqual(z, 0)
            self.assertEqual(err, 0)

            hop.set_position(i, x, i*i, 0)

        for i in range(6):
            x, y, z, err = hop.get_position(i)
            self.assertEqual(x, y)
            self.assertEqual(z, 0)
            self.assertEqual(err, 0)

        hop.stop()

    def test2(self):
        random.seed(1001)

        hop = HopInterface()
        hop.initialize_code()

        particles = new_plummer_model(1000)
        ids, errors = hop.new_particle(
            particles.mass.value_in(nbody_system.mass),
            particles.x.value_in(nbody_system.length),
            particles.y.value_in(nbody_system.length),
            particles.z.value_in(nbody_system.length)
        )

        n, err = hop.get_number_of_particles()
        self.assertEqual(n, 1000)
        self.assertEqual(err, 0)

        # distance_to_center = (particles.position - particles.center_of_mass()).lengths()

        # print distance_to_center
        ds = {0: 0.482308834791, 1: 0.4885137677192688, 2: 0.27442726492881775}
        for method in [0, 1, 2]:
            hop.set_nDens(7)
            hop.set_density_method(method)
            hop.calculate_densities()

            d, err = hop.get_density(0)
            self.assertAlmostRelativeEquals(d, ds[method], 5)
        hop.stop()

    def test3(self):

        random.seed(1001)

        print("Third test: densest neighbors and groups.")

        hop = HopInterface()
        hop.initialize_code()

        particles1 = new_plummer_model(10)
        particles2 = new_plummer_model(10)
        particles3 = new_plummer_model(10)

        particles2.position += (10, 0, 0) | nbody_system.length

        particles3.position += (0, 20, 0) | nbody_system.length

        ids1, errors = hop.new_particle(
            particles1.mass.value_in(nbody_system.mass),
            particles1.x.value_in(nbody_system.length),
            particles1.y.value_in(nbody_system.length),
            particles1.z.value_in(nbody_system.length)
        )

        ids2, errors = hop.new_particle(
            particles2.mass.value_in(nbody_system.mass),
            particles2.x.value_in(nbody_system.length),
            particles2.y.value_in(nbody_system.length),
            particles2.z.value_in(nbody_system.length)
        )

        ids3, errors = hop.new_particle(
            particles3.mass.value_in(nbody_system.mass),
            particles3.x.value_in(nbody_system.length),
            particles3.y.value_in(nbody_system.length),
            particles3.z.value_in(nbody_system.length)
        )

        hop.set_nDens(5)
        hop.calculate_densities()
        hop.do_hop()

        n, err = hop.get_group_id(ids1)
        self.assertEqual(n, 2)
        self.assertEqual(err, 0)
        n, err = hop.get_group_id(ids2)
        self.assertEqual(err, 0)
        n, err = hop.get_group_id(ids3)
        self.assertEqual(err, 0)

        n, err = hop.get_densest_particle_in_group(2)
        self.assertEqual(n, 7)
        for i in range(3):
            n, err = hop.get_number_of_particles_in_group(0)
            self.assertEqual(err, 0)
            self.assertEqual(n, 10)

        n, err = hop.get_number_of_groups()
        self.assertEqual(n, 3)

        n, err = hop.get_densest_neighbor(ids1)
        self.assertEqual(n, [7, 7, 12, 0, 7, 7, 7, 7, 12, 7])
        hop.stop()

    def test4(self):
        hop = HopInterface()
        hop.initialize_code()
        value, error = hop.get_nDens()
        self.assertEqual(error, 0)
        self.assertEqual(value, 64)
        error = hop.set_nDens(7)
        self.assertEqual(error, 0)
        value, error = hop.get_nDens()
        self.assertEqual(value, 7)

        value, error = hop.get_nHop()
        self.assertEqual(error, 0)
        self.assertEqual(value, -1)
        error = hop.set_nHop(7)
        self.assertEqual(error, 0)
        value, error = hop.get_nHop()
        self.assertEqual(value, 7)

        value, error = hop.get_nBucket()
        self.assertEqual(error, 0)
        self.assertEqual(value, 16)
        error = hop.set_nHop(7)
        self.assertEqual(error, 0)
        value, error = hop.get_nHop()
        self.assertEqual(value, 7)
        hop.stop()


class TestHop(TestCase):
    def test1(self):
        print("First test: adding particles, setting and getting.")
        hop = Hop()
        particles = Particles(6)
        particles.mass = 1.0 | nbody_system.mass
        particles.x = [i*i for i in range(6)] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length

        hop.particles.add_particles(particles)

        positions = hop.particles.position
        for i in range(6):
            x, y, z = positions[i]
            self.assertEqual(x, i*i | nbody_system.length)
            self.assertEqual(y, 0 | nbody_system.length)
            self.assertEqual(z, 0 | nbody_system.length)

        hop.stop()

    def test2(self):
        random.seed(1001)

        hop = Hop()
        hop.initialize_code()
        hop.parameters.number_of_neighbors_for_local_density = 7
        hop.commit_parameters()

        particles = new_plummer_model(1000)
        hop.particles.add_particles(particles)

        # distance_to_center = (particles.position - particles.center_of_mass()).lengths()

        # print distance_to_center
        ds = {0: 0.482308834791, 1: 0.4885137677192688, 2: 0.27442726492881775}
        for method in [0, 1, 2]:
            hop.set_density_method(method)
            hop.calculate_densities()

            d = hop.particles[0].density

            self.assertAlmostRelativeEquals(d, ds[method] | nbody_system.density, 5)
        hop.stop()

    def test3(self):

        random.seed(1001)

        print("Third test: densest neighbors and groups.")

        hop = Hop()
        hop.parameters.number_of_neighbors_for_local_density = 5

        particles1 = new_plummer_model(10)
        particles2 = new_plummer_model(10)
        particles3 = new_plummer_model(10)

        particles2.position += (10, 0, 0) | nbody_system.length

        particles3.position += (0, 20, 0) | nbody_system.length

        hop.particles.add_particles(particles1)
        hop.particles.add_particles(particles2)
        hop.particles.add_particles(particles3)

        hop.calculate_densities()
        hop.do_hop()

        print(hop.particles.group_id)

        groups = list(hop.groups())

        self.assertEqual(len(groups), 3)

        self.assertEqual(hop.get_number_of_particles_outside_groups(), 0)

        # densities = (0,0,0) | nbody_system.density
        for index, group in enumerate(groups):
            self.assertEqual(len(group), 10)
            self.assertEqual(group.id_of_group(), index)
            # self.assertEquals(group.get_density_of_group(), densities[index])
        hop.stop()

    def test4(self):
        random.seed(1001)
        print("Test 4: complicated density field.")

        # A separate group below peak_density_threshold -> should be dropped
        particles0 = new_plummer_model(90, convert_nbody=nbody_system.nbody_to_si(0.9 | units.MSun, 1.0 | units.RSun))

        # A nearby group below peak_density_threshold -> should be attached to proper group
        particles1 = new_plummer_model(80, convert_nbody=nbody_system.nbody_to_si(0.8 | units.MSun, 1.0 | units.RSun))
        particles1.x += 10 | units.RSun

        # A proper group very nearby other proper group -> groups should merge
        particles2a = new_plummer_model(200, convert_nbody=nbody_system.nbody_to_si(2.0 | units.MSun, 1.0 | units.RSun))
        particles2b = new_plummer_model(300, convert_nbody=nbody_system.nbody_to_si(3.0 | units.MSun, 1.0 | units.RSun))
        particles2a.x += 11.0 | units.RSun
        particles2b.x += 11.2 | units.RSun

        # A separate proper group other proper group -> groups should be preserved
        particles3 = new_plummer_model(400, convert_nbody=nbody_system.nbody_to_si(4.0 | units.MSun, 1.0 | units.RSun))
        particles3.x += 20 | units.RSun

        hop = Hop(unit_converter=nbody_system.nbody_to_si(10.7 | units.MSun, 1.0 | units.RSun))
        hop.parameters.number_of_neighbors_for_local_density = 100
        hop.parameters.saddle_density_threshold_factor = 0.5
        hop.parameters.relative_saddle_density_threshold = True
        hop.commit_parameters()

        for set in [particles0, particles1, particles2a, particles2b, particles3]:
            hop.particles.add_particles(set)

        hop.calculate_densities()
        hop.parameters.outer_density_threshold = 0.1 * hop.particles.density.mean()
        hop.parameters.peak_density_threshold = hop.particles.density.amax() / 4.0
        hop.recommit_parameters()
        hop.do_hop()
        groups = list(hop.groups())

        self.assertEqual(len(hop.particles), 1070)
        self.assertEqual(len(groups), 2)
        self.assertEqual(hop.particles.select(lambda x: x < 5 | units.RSun, "x").group_id, -1)
        self.assertEqual(hop.get_number_of_particles_outside_groups(), 299)
        self.assertEqual(1070 - len(groups[0]) - len(groups[1]), 299)

        expected_size = [477, 294]  # Less than [580, 400], because particles below outer_density_threshold are excluded
        expected_average_x = [11.0, 20] | units.RSun
        for index, group in enumerate(groups):
            self.assertEqual(group.id_of_group(), index)
            self.assertAlmostEqual(group.center_of_mass()[0], expected_average_x[index], 1)
            self.assertEqual(len(group), expected_size[index])

        if False:  # Make a plot
            original = hop.particles.copy()
            from amuse.plot import scatter, native_plot
            colors = ["r", "g", "b", "y", "k", "w"]*100
            for group, color in zip(hop.groups(), colors):
                scatter(group.x, group.y, c=color)
                original -= group
            scatter(original.x, original.y, c="m", marker="s")
            native_plot.show()

        hop.stop()

    def test5(self):
        print("Test error codes")
        unit_converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        particles = new_plummer_model(200, convert_nbody=unit_converter)
        hop = Hop(unit_converter=unit_converter)  # , redirection="none")
        hop.parameters.number_of_neighbors_for_local_density = 100
        hop.particles.add_particles(particles[:99])
        self.assertRaises(AmuseException, hop.calculate_densities, expected_message="Error when calling 'calculate_densities' of a 'Hop', errorcode is -5, error is 'Too few particles.'")
        hop.particles.add_particles(particles[99:101])
        hop.calculate_densities()
        hop.parameters.number_of_neighbors_for_hop = 200
        self.assertRaises(AmuseException, hop.calculate_densities, expected_message="Error when calling 'calculate_densities' of a 'Hop', errorcode is -5, error is 'Too few particles.'")
        hop.particles.add_particles(particles[101:])
        hop.calculate_densities()

        self.assertRaises(AmuseException, hop.get_mass, 200, expected_message="Error when calling 'get_mass' of a 'Hop', errorcode is -3, error is 'A particle with the given index was not found.'")
        hop.stop()

    def test6(self):
        print("Test with different masses")
        # Particles on a cubic grid with masses according to a gaussian density profile
        grid = numpy.mgrid[-1:1:21j, -1:1:21j, -1:1:21j] | units.m
        particles = Particles(9261, x=grid[0].flatten(), y=grid[1].flatten(), z=grid[2].flatten())
        peak_positions = [[0.2, -0.4, 0.3], [-0.6, 0.2, 0.7]] | units.m
        particles.mass = 2*numpy.exp(-(particles.position-peak_positions[0]).lengths_squared() / (0.1 | units.m**2)) | units.kg
        particles.mass += numpy.exp(-(particles.position-peak_positions[1]).lengths_squared() / (0.1 | units.m**2)) | units.kg
        self.assertAlmostEqual(particles.position[particles.mass.argmax()], peak_positions[0])
        self.assertAlmostEqual(particles[:4000].position[particles[:4000].mass.argmax()], peak_positions[1])

        hop = Hop(unit_converter=nbody_system.nbody_to_si(particles.mass.sum(), 1.0 | units.m))  # , redirection="none")
        hop.parameters.density_method = 2
        hop.parameters.number_of_neighbors_for_local_density = 50
        hop.parameters.relative_saddle_density_threshold = True
        hop.commit_parameters()
        hop.particles.add_particles(particles)
        hop.calculate_densities()
        self.assertAlmostEqual(hop.particles.position[hop.particles.density.argmax()], peak_positions[0])
        self.assertAlmostEqual(hop.particles[:4000].position[hop.particles[:4000].density.argmax()], peak_positions[1])
        hop.do_hop()
        groups = list(hop.groups())
        self.assertEqual(len(groups), 2)
        for group, peak_position in zip(groups, peak_positions):
            self.assertAlmostEqual(group.center_of_mass(), peak_position, 1)
        hop.stop()

    def test7(self):
        print("Testing Hop states")
        unit_converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        particles = new_plummer_model(200, convert_nbody=unit_converter)

        print("First do everything manually:", end=' ')
        instance = Hop(unit_converter=unit_converter)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particles(particles)
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()
        print("ok")

        print("initialize_code(), commit_parameters(), (re)commit_particles(), "
            "and cleanup_code() should be called automatically:", end=' ')
        instance = Hop(unit_converter=unit_converter)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.number_of_neighbors_for_local_density = 50
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particles(particles[:100])
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.add_particles(particles[100:])
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.particles[100].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')
        print("ok")

    def test8(self):
        random.seed(1001)
        print("Test 8: SI vs nbody units.")
        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.RSun)
        numpy.random.seed(1234)
        particles = new_plummer_model(1000, convert_nbody=converter)
        hop = Hop(unit_converter=converter)
        hop.particles.add_particles(particles)
        hop.calculate_densities()
        hop.parameters.outer_density_threshold = 0.1 | nbody_system.mass / nbody_system.length**3
        hop.do_hop()
        groups = list(hop.groups())

        self.assertEqual(len(hop.particles), 1000)
        self.assertEqual(len(groups), 1)
        self.assertEqual(len(groups[0]), 511)
        self.assertEqual(len(hop.no_group()), 489)
        hop.stop()

        numpy.random.seed(1234)
        particles = new_plummer_model(1000)
        hop = Hop()
        hop.particles.add_particles(particles)
        hop.calculate_densities()
        hop.parameters.outer_density_threshold = 0.1 | nbody_system.mass / nbody_system.length**3
        hop.do_hop()
        groups = list(hop.groups())

        self.assertEqual(len(hop.particles), 1000)
        self.assertEqual(len(groups), 1)
        self.assertEqual(len(groups[0]), 511)
        self.assertEqual(len(hop.no_group()), 489)
        hop.stop()



import math
import numpy
import os.path

from amuse.test.amusetest import TestWithMPI
from amuse.units import nbody_system, units, constants
from amuse.datamodel import Particles, Particle
from amuse.community.pikachu.interface import PikachuInterface, Pikachu

from amuse.ic.plummer import new_plummer_model


default_options = dict(mode='normal')
# ~default_options = dict(redirection="none", mode='normal')
# ~default_options = dict(debugger="gdb", mode='normal')
# ~default_options = dict(redirection="none", mode='large_n')


class TestPikachuInterface(TestWithMPI):

    def test1(self):
        print("Test PikachuInterface initialization")
        instance = self.new_instance_of_an_optional_code(PikachuInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())

        directory, error = instance.get_kernel_directory()
        self.assertEqual(0, error)
        self.assertEqual("./", directory)
        self.assertEqual(0, instance.set_kernel_directory(instance.default_kernel_directory))
        directory, error = instance.get_kernel_directory()
        self.assertEqual(error, 0)
        self.assertEqual(directory, os.path.join(instance.amuse_root_directory, "src", "amuse", "community", "pikachu"))

        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test2(self):
        print("Test PikachuInterface new_particle / get_state")
        instance = self.new_instance_of_an_optional_code(PikachuInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_kernel_directory(instance.default_kernel_directory))
        self.assertEqual(0, instance.commit_parameters())

        id, error = instance.new_particle(mass=11.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0, radius=1.0)
        self.assertEqual(0, error)
        self.assertEqual(1, id)
        id, error = instance.new_particle(mass=21.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0, radius=2.0)
        self.assertEqual(0, error)
        self.assertEqual(2, id)
        self.assertEqual(0, instance.commit_particles())
        retrieved_state1 = instance.get_state(1)
        retrieved_state2 = instance.get_state(2)
        self.assertEqual(0,  retrieved_state1['__result'])
        self.assertEqual(0,  retrieved_state2['__result'])
        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test3(self):
        print("Test PikachuInterface particle property getters/setters")
        instance = self.new_instance_of_an_optional_code(PikachuInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_kernel_directory(instance.default_kernel_directory))
        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual([1, 0], list(instance.new_particle(0.01,  1, 0, 0,  0, 1, 0, 0.1).values()))
        self.assertEqual([2, 0], list(instance.new_particle(0.02, -1, 0, 0,  0, -1, 0, 0.1).values()))
        self.assertEqual(-3, instance.get_mass(1)['__result'])  # Have to commit first
        self.assertEqual(0, instance.commit_particles())

        # getters
        mass, result = instance.get_mass(1)
        self.assertAlmostEqual(0.01, mass)
        self.assertEqual(0, result)
        radius, result = instance.get_radius(2)
        self.assertAlmostEqual(0.1, radius)
        self.assertEqual(0, result)
        self.assertEqual(-3, instance.get_mass(3)['__result'])  # Particle not found
        self.assertEqual([1, 0, 0,  0], list(instance.get_position(1).values()))
        self.assertEqual([-1, 0, 0,  0], list(instance.get_position(2).values()))
        self.assertEqual([0, 1, 0,  0], list(instance.get_velocity(1).values()))
        self.assertEqual([0, -1, 0,  0], list(instance.get_velocity(2).values()))

        # setters
        self.assertEqual(0, instance.set_state(1, 0.01, 1, 2, 3, 4, 5, 6, 0.1))
        self.assertEqual([0.01, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.1, 0], list(instance.get_state(1).values()))
        self.assertEqual(0, instance.set_mass(1, 0.02))
        self.assertEqual([0.02, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.1, 0], list(instance.get_state(1).values()))
        self.assertEqual(0, instance.set_radius(1, 0.2))
        self.assertEqual([0.02, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 0.2, 0], list(instance.get_state(1).values()))
        self.assertEqual(0, instance.set_position(1, 10, 20, 30))
        self.assertEqual([0.02, 10.0, 20.0, 30.0, 4.0, 5.0, 6.0, 0.2, 0], list(instance.get_state(1).values()))
        self.assertEqual(0, instance.set_velocity(1, 40, 50, 60))
        self.assertEqual([0.02, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 0.2, 0], list(instance.get_state(1).values()))

        self.assertEqual(0, instance.cleanup_code())
        instance.stop()

    def test4(self):
        print("Test PikachuInterface parameters")
        instance = self.new_instance_of_an_optional_code(PikachuInterface, **default_options)
        self.assertEqual(0, instance.initialize_code())

        self.assertEqual(0, instance.set_kernel_directory(instance.default_kernel_directory))

        # Pikachu has separate epsilon_squared parameters for different interactions!
        self.assertEqual([1.0e-8, 0], list(instance.get_eps2_fs_fs().values()))
        self.assertEqual([1.0e-8, 0], list(instance.get_eps2_fs_bh().values()))
        self.assertEqual([0, 0], list(instance.get_eps2_bh_bh().values()))
        self.assertEqual(-2, instance.get_eps2()['__result'])  # Not implemented (would be ambiguous)

        self.assertEqual(0, instance.set_eps2_fs_fs(0.2))
        self.assertEqual([0.2, 0], list(instance.get_eps2_fs_fs().values()))
        self.assertEqual(0, instance.set_eps2_fs_bh(0.3))
        self.assertEqual([0.3, 0], list(instance.get_eps2_fs_bh().values()))
        self.assertEqual(0, instance.set_eps2_bh_bh(0.4))
        self.assertEqual([0.4, 0], list(instance.get_eps2_bh_bh().values()))
        self.assertEqual(-2, instance.set_eps2(0.1))  # Not implemented (would be ambiguous)

        self.assertEqual([0.005, 0], list(instance.get_eta_s().values()))
        self.assertEqual([0.025, 0], list(instance.get_eta_fs().values()))
        self.assertEqual([0.025, 0], list(instance.get_eta_smbh().values()))

        self.assertEqual(0, instance.set_eta_s(0.01))
        self.assertEqual([0.01, 0], list(instance.get_eta_s().values()))
        self.assertEqual(0, instance.set_eta_fs(0.02))
        self.assertEqual([0.02, 0], list(instance.get_eta_fs().values()))
        self.assertEqual(0, instance.set_eta_smbh(0.03))
        self.assertEqual([0.03, 0], list(instance.get_eta_smbh().values()))

        self.assertEqual(0, instance.commit_parameters())
        self.assertEqual(0, instance.cleanup_code())
        instance.stop()


class TestPikachu(TestWithMPI):

    default_converter = nbody_system.nbody_to_si(1.0e4 | units.MSun, 1.0 | units.AU)

    def new_sun_earth_system(self):
        particles = Particles(2)
        particles.mass = [1.0, 3.0037e-6] | units.MSun
        particles.position = [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.km / units.s
        particles[1].vy = (constants.G * particles.total_mass() / (1.0 | units.AU)).sqrt()
        return particles

    def test1(self):
        print("Testing Pikachu initialization")
        instance = self.new_instance_of_an_optional_code(Pikachu, self.default_converter, **default_options)
        instance.initialize_code()
        instance.commit_parameters()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Testing Pikachu parameters")
        instance = self.new_instance_of_an_optional_code(Pikachu, self.default_converter, **default_options)
        instance.initialize_code()

        self.assertEqual(instance.parameters.epsilon_squared,
            instance.unit_converter.to_si(1.0e-8 | nbody_system.length**2))
        self.assertEqual(instance.parameters.timestep_parameter, 0.025)

        for par, value in [('epsilon_squared_star_star', 1.0e-8 | nbody_system.length**2),
                ('epsilon_squared_star_blackhole', 1.0e-8 | nbody_system.length**2),
                ('epsilon_squared_blackhole_blackhole', 0.0 | nbody_system.length**2),
                ('initial_timestep_parameter', 0.005),
                ('timestep_parameter_stars', 0.025),
                ('timestep_parameter_black_holes', 0.025),
                ('timestep', 1.0/2048.0 | nbody_system.time),
                ('search_factor', 3.0),
                ('velocity_dispersion', 0.707106781 | nbody_system.speed),
                ('rcut_out_star_star', 2.0e-3 | nbody_system.length),
                ('rcut_out_star_blackhole', 2.0e-2 | nbody_system.length),
                ('rcut_out_blackhole_blackhole', 1.0e5 | nbody_system.length),
                ('rsearch_star_star', 0.0 | nbody_system.length),
                ('rsearch_star_blackhole', 0.0 | nbody_system.length),
                ('rsearch_blackhole_blackhole', 0.0 | nbody_system.length),
                ('opening_angle', 0.4)]:
            self.assertEqual(instance.unit_converter.to_si(value),
                getattr(instance.parameters, par))

            if hasattr(value, 'unit'):
                new_value = 3.0 | value.unit
            else:
                new_value = 3.0

            setattr(instance.parameters, par, new_value)
            self.assertEqual(instance.unit_converter.to_si(new_value),
                getattr(instance.parameters, par))

        # epsilon_squared is an alias for epsilon_squared_star_star, so epsilon_squared also has become 3:
        self.assertEqual(instance.parameters.epsilon_squared,
            instance.unit_converter.to_si(3.0 | nbody_system.length**2))
        instance.parameters.epsilon_squared = 0.1 | nbody_system.length**2
        self.assertEqual(instance.parameters.epsilon_squared,
            instance.unit_converter.to_si(0.1 | nbody_system.length**2))
        # timestep_parameter is an alias for timestep_parameter_stars, so timestep_parameter also has become 3:
        self.assertEqual(instance.parameters.timestep_parameter, 3.0)
        instance.parameters.timestep_parameter = 0.01
        self.assertEqual(instance.parameters.timestep_parameter, 0.01)

        self.assertEqual(instance.parameters.calculate_quadrupole_moments, False)
        instance.parameters.calculate_quadrupole_moments = True
        self.assertEqual(instance.parameters.calculate_quadrupole_moments, True)

        instance.commit_parameters()
        p = instance.parameters
        self.assertAlmostRelativeEquals(p.rsearch_star_star,
            p.rcut_out_star_star + p.search_factor * p.velocity_dispersion * p.timestep, 10)
        self.assertAlmostRelativeEquals(p.rsearch_star_blackhole,
            p.rcut_out_star_blackhole + p.search_factor * p.velocity_dispersion * p.timestep, 10)
        self.assertAlmostRelativeEquals(p.rsearch_blackhole_blackhole,
            p.rcut_out_blackhole_blackhole + p.search_factor * p.velocity_dispersion * p.timestep, 10)

        instance.stop()

    def test3(self):
        print("Testing Pikachu particles")
        instance = self.new_instance_of_an_optional_code(Pikachu, self.default_converter, **default_options)
        instance.initialize_code()
        instance.commit_parameters()
        instance.particles.add_particles(self.new_sun_earth_system())
        instance.commit_particles()

        self.assertAlmostEqual(instance.particles.mass, [1.0, 3.0037e-6] | units.MSun)
        self.assertAlmostEqual(instance.particles.position,
            [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]] | units.AU)
        self.assertAlmostEqual(instance.particles.velocity,
            [[0.0, 0.0, 0.0], [0.0, 29.7885, 0.0]] | units.km / units.s, 3)

        instance.cleanup_code()
        instance.stop()

    def test4(self):
        print("Testing Pikachu evolve_model, 2 particles")
        particles = Particles(2)
        particles.mass = 1.0 | units.MSun
        particles.radius = 1.0 | units.RSun
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | units.AU
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.km / units.s
        particles[1].vy = (constants.G * (2.0 | units.MSun) / (2.0 | units.AU)).sqrt()
        particles.move_to_center()

        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Pikachu, converter, **default_options)
        instance.initialize_code()
        instance.parameters.timestep = 0.0125 * math.pi * particles[0].x / particles[0].vy
        instance.parameters.rcut_out_star_star = 10.0 | units.AU
        instance.commit_parameters()
        instance.particles.add_particles(particles)
        instance.commit_particles()
        primary = instance.particles[0]

        P = 2 * math.pi * primary.x / primary.vy

        position_at_start = primary.position.x
        instance.evolve_model(P / 4.0)
        self.assertAlmostRelativeEqual(position_at_start, primary.position.y, 3)

        instance.evolve_model(P / 2.0)
        self.assertAlmostRelativeEqual(position_at_start, -primary.position.x, 3)

        instance.evolve_model(P)
        self.assertAlmostRelativeEqual(position_at_start, primary.position.x, 3)

        instance.cleanup_code()
        instance.stop()

    def test5(self):
        print("Testing Pikachu evolve_model, earth-sun system")
        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        instance = self.new_instance_of_an_optional_code(Pikachu, converter, **default_options)
        instance.initialize_code()
        instance.parameters.timestep = 0.005 | units.yr
        instance.parameters.rcut_out_star_star = 1.0 | units.AU
        instance.commit_parameters()
        instance.particles.add_particles(self.new_sun_earth_system())
        instance.commit_particles()
        earth = instance.particles[1]

        position_at_start = earth.position.x
        instance.evolve_model(0.25 | units.yr)
        self.assertAlmostRelativeEqual(position_at_start, earth.position.y, 3)

        instance.evolve_model(0.5 | units.yr)
        self.assertAlmostRelativeEqual(position_at_start, -earth.position.x, 3)

        instance.evolve_model(1.0 | units.yr)
        self.assertAlmostRelativeEqual(position_at_start, earth.position.x, 3)

        instance.cleanup_code()
        instance.stop()

    def test6a(self):
        print("Testing effect of Pikachu parameter epsilon_squared (using recommit_particles)")
        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        particles = self.new_sun_earth_system()
        particles.rotate(0.0, 0.0, math.pi/4)
        particles.move_to_center()

        tan_initial_direction = particles[1].vy/particles[1].vx
        self.assertAlmostEqual(tan_initial_direction, math.tan(-math.pi/4))
        tan_final_direction = []

        instance = self.new_instance_of_an_optional_code(Pikachu, converter, **default_options)
        instance.initialize_code()
        instance.parameters.timestep = 0.005 | units.yr

        for log_eps2 in range(-5, 6, 2):
            instance.parameters.epsilon_squared = 10.0**log_eps2 | units.AU ** 2
            if len(instance.particles) > 0:
                instance.particles.remove_particles(instance.particles)
            instance.particles.add_particles(particles)
            instance.evolve_model(instance.model_time.as_quantity_in(units.yr) + (0.25 | units.yr))
            tan_final_direction.append(instance.particles[1].velocity[1] /
                instance.particles[1].velocity[0])
        print(tan_final_direction)
        # Small values of epsilon_squared should result in normal earth-sun dynamics: rotation of 90 degrees
        self.assertAlmostRelativeEquals(tan_final_direction[0], math.tan(math.pi / 4.0), 1)
        # Large values of epsilon_squared should result in ~ no interaction
        self.assertAlmostEqual(tan_final_direction[-1], tan_initial_direction, 2)
        # Outcome is most sensitive to epsilon_squared when epsilon_squared = d(earth, sun)^2
        delta = [abs(tan_final_direction[i+1]-tan_final_direction[i]) for i in range(len(tan_final_direction)-1)]
        self.assertEqual(delta[len(tan_final_direction)/2 - 1], max(delta))

    def test6b(self):
        print("Testing effect of Pikachu parameter epsilon_squared (using reset)")
        converter = nbody_system.nbody_to_si(1.0 | units.MSun, 1.0 | units.AU)
        particles = self.new_sun_earth_system()
        particles.rotate(0.0, 0.0, math.pi/4)
        particles.move_to_center()

        tan_initial_direction = particles[1].vy/particles[1].vx
        self.assertAlmostEqual(tan_initial_direction, math.tan(-math.pi/4))
        tan_final_direction = []

        instance = self.new_instance_of_an_optional_code(Pikachu, converter, **default_options)
        instance.initialize_code()
        instance.parameters.timestep = 0.005 | units.yr
        for log_eps2 in range(-5, 6, 2):
            instance.parameters.epsilon_squared = 10.0**log_eps2 | units.AU ** 2
            instance.commit_parameters()
            instance.particles.add_particles(particles)
            instance.commit_particles()
            instance.evolve_model(0.25 | units.yr)
            tan_final_direction.append(instance.particles[1].velocity[1] /
                instance.particles[1].velocity[0])
            instance.reset()
        instance.cleanup_code()
        instance.stop()
        print(tan_final_direction)
        # Small values of epsilon_squared should result in normal earth-sun dynamics: rotation of 90 degrees
        self.assertAlmostRelativeEquals(tan_final_direction[0], math.tan(math.pi / 4.0), 1)
        # Large values of epsilon_squared should result in ~ no interaction
        self.assertAlmostEqual(tan_final_direction[-1], tan_initial_direction, 2)
        # Outcome is most sensitive to epsilon_squared when epsilon_squared = d(earth, sun)^2
        delta = [abs(tan_final_direction[i+1]-tan_final_direction[i]) for i in range(len(tan_final_direction)-1)]
        self.assertEqual(delta[len(tan_final_direction)/2 - 1], max(delta))

    def test7(self):
        print("Testing Pikachu states")
        stars = new_plummer_model(100)
        black_hole = Particle()
        black_hole.mass = 1.0 | nbody_system.mass
        black_hole.radius = 0.0 | nbody_system.length
        black_hole.position = [0.0, 0.0, 0.0] | nbody_system.length
        black_hole.velocity = [0.0, 0.0, 0.0] | nbody_system.speed

        print("First do everything manually:")
        instance = self.new_instance_of_an_optional_code(Pikachu, **default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.initialize_code()
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
# ~        instance.parameters.rcut_out_star_star = 1.0 | nbody_system.length
        instance.parameters.timestep = 0.001 | nbody_system.time
        instance.commit_parameters()
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        instance.particles.add_particles(stars)
        instance.commit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.remove_particle(stars[0])
        instance.particles.add_particle(black_hole)
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        instance.recommit_particles()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.evolve_model(0.001 | nbody_system.time)
        self.assertEqual(instance.get_name_of_current_state(), 'EVOLVED')
        instance.synchronize_model()
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.cleanup_code()
        self.assertEqual(instance.get_name_of_current_state(), 'END')
        instance.stop()

        print("initialize_code(), commit_parameters(), (re)commit_particles(), "
            "synchronize_model(), and cleanup_code() should be called "
            "automatically before editing parameters, new_particle(), get_xx(), and stop():")
        instance = self.new_instance_of_an_optional_code(Pikachu, **default_options)
        self.assertEqual(instance.get_name_of_current_state(), 'UNINITIALIZED')
        instance.parameters.timestep = 0.001 | nbody_system.time
        self.assertEqual(instance.get_name_of_current_state(), 'INITIALIZED')
        instance.particles.add_particles(stars)
        self.assertEqual(instance.get_name_of_current_state(), 'EDIT')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.particles.remove_particle(stars[0])
        instance.particles.add_particle(black_hole)
        self.assertEqual(instance.get_name_of_current_state(), 'UPDATE')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.evolve_model(0.001 | nbody_system.time)
        self.assertEqual(instance.get_name_of_current_state(), 'EVOLVED')
        mass = instance.particles[0].mass
        self.assertEqual(instance.get_name_of_current_state(), 'RUN')
        instance.stop()
        self.assertEqual(instance.get_name_of_current_state(), 'STOPPED')



import unittest
import docutils.core

from amuse.support import literature


class TestLiteratureRefs(unittest.TestCase):
    def setUp(self):
        literature.LiteratureReferencesMixIn.literature_list = []

    def test1(self):
        class ClassLitrefs(literature.LiteratureReferencesMixIn):
            """ some comments with added lit refs, i.e.  [#]_ and [3]_ etc...

                .. [#] Gratia, Exempli, *Journal of I.M.*, **1**, 1--100 (2009).
                .. [3] Dude, John, *The Intern. Foo Journal of Bars*, **51**, 1647--1751 (2009)  
                .. [4] Hat, John, *The Ex. Foo Journal of Bars*, **51**, 1647--1751 (2009)  
            """

            def __init__(self):
                literature.LiteratureReferencesMixIn.__init__(self)

        classnames = literature.LiteratureReferencesMixIn.names_of_classes_with_references()
        for classname in classnames:
            self.assertFalse("ClassLitrefs" in classname)

        instance = ClassLitrefs()

        classnames = literature.LiteratureReferencesMixIn.names_of_classes_with_references()
        for classname in classnames:
            self.assertTrue("ClassLitrefs" in classname)

    def test2(self):
        class ClassLitrefs(literature.LiteratureReferencesMixIn):
            """ some comments with added lit refs, i.e.  [#]_ and [3]_ etc...

                .. [#] Gratia, Exempli, *Journal of I.M.*, **1**, 1--100 (2009).
                .. [3] Dude, John, *The Intern. Foo Journal of Bars*, **51**, 1647--1751 (2009)  
                .. [4] Hat, John, *The Gal. Foo Journal of Bars*, **3**, 16--51 (2009)  
            """

            def __init__(self):
                literature.LiteratureReferencesMixIn.__init__(self)

        literature.TrackLiteratureReferences.default().registered_classes = set([])
        string = literature.LiteratureReferencesMixIn.all_literature_references_string()
        self.assertTrue("AMUSE" in string)
        self.assertTrue("2018araa.book.....P" in string)

        instance = ClassLitrefs()

        string = literature.LiteratureReferencesMixIn.all_literature_references_string()
        literature.LiteratureReferencesMixIn.print_literature_references()
        self.assertFalse(".. [#] Gratia, Exem" in string)
        self.assertFalse(".. [3] Dude" in string)
        self.assertFalse(".. [4] Hat" in string)

        self.assertTrue("Gratia, Exem" in string)
        self.assertTrue("Dude" in string)
        self.assertTrue("Hat" in string)

        self.assertTrue("ClassLitrefs" in string)



import numpy

from amuse.test.amusetest import TestCase
from amuse.support.exceptions import AmuseWarning, AmuseException
from amuse.units import units, nbody_system, generic_unit_system, quantities
from amuse.units.generic_unit_converter import ConvertBetweenGenericAndSiUnits
from amuse.units.quantities import zero, VectorQuantity
from amuse.datamodel import Particle, Particles, ParticlesSuperset
from amuse.community.interface.gd import GravitationalDynamics
from amuse.community.interface.stopping_conditions import StoppingConditions
from amuse.ext.sink import SinkParticles, new_sink_particles
from amuse.ext import sink

from amuse.community.gadget2.interface import Gadget2
from amuse.ext.evrard_test import new_evrard_gas_sphere
from amuse.ic.plummer import new_plummer_model


class TestSinkParticles(TestCase):

    looping_over = "sinks"

    def test1(self):
        print("Testing SinkParticles initialization from new (blank) particle")
        sinks = SinkParticles(Particles(2), sink_radius=[42.0, 43.0] | units.RSun)

        self.assertEqual(sinks.sink_radius, [42.0, 43.0] | units.RSun)
        self.assertEqual(sinks.mass, 0.0 | units.MSun)
        self.assertEqual(sinks.position, [0.0, 0.0, 0.0] | units.parsec)
        self.assertEqual(sinks.velocity, [0.0, 0.0, 0.0] | units.km/units.s)
        self.assertEqual(sinks.angular_momentum, [0.0, 0.0, 0.0] | units.kg*units.km**2/units.s)

        sinks = SinkParticles(Particles(2), sink_radius=24.0 | units.RSun, mass=[1.0, 2.0] | units.MSun,
            position=[1, 2, 3] | nbody_system.length, velocity=[4, 5, 6] | nbody_system.length/nbody_system.time,
            angular_momentum=[9, 8, 7] | nbody_system.mass*nbody_system.length**2/nbody_system.time)
        self.assertEqual(sinks.sink_radius, 24.0 | units.RSun)
        self.assertEqual(sinks.mass, [1.0, 2.0] | units.MSun)
        self.assertEqual(sinks.position, [1.0, 2.0, 3.0] | nbody_system.length)
        self.assertEqual(sinks.velocity, [4.0, 5.0, 6.0] | nbody_system.length/nbody_system.time)
        self.assertEqual(sinks.angular_momentum, [9.0, 8.0, 7.0] | nbody_system.mass*nbody_system.length**2/nbody_system.time)

    def test2(self):
        print("Testing SinkParticles initialization from existing particles")
        original = Particles(3)
        self.assertRaises(AttributeError, SinkParticles, original, expected_message="You tried to access attribute 'radius' but this attribute is not defined for this set.")
        original.radius = 42.0 | units.RSun
        original.mass = 10.0 | units.MSun
        original.position = [[i, -i, 2*i] for i in range(3)] | units.parsec
        sinks = SinkParticles(original)
        self.assertEqual(sinks.sink_radius, 42.0 | units.RSun)
        self.assertEqual(sinks.mass, 10.0 | units.MSun)
        self.assertEqual(sinks.position, [[0, 0, 0], [1, -1, 2], [2, -2, 4]] | units.parsec)
        self.assertRaises(AttributeError, getattr, sinks, "bogus", expected_message="You tried to access attribute 'bogus' but this attribute is not defined for this set.")

    def test3(self):
        print("Testing SinkParticles initialization from existing particles in set")
        particles = Particles(10)
        self.assertRaises(AttributeError, SinkParticles, particles[[4, 7]], expected_message="You tried to access attribute 'radius' but this attribute is not defined for this set.")
        particles.radius = 42.0 | units.RSun
        particles.mass = range(1, 11) | units.MSun
        particles.position = [[i, 2*i, 3*i] for i in range(10)] | units.parsec

        sinks = SinkParticles(particles[[4]])
        self.assertEqual(sinks.mass, 5.0 | units.MSun)
        self.assertEqual(sinks.sink_radius, 42.0 | units.RSun)
        self.assertEqual(sinks.radius, 42.0 | units.RSun)
        self.assertEqual(sinks.position, [4.0, 8.0, 12.0] | units.parsec)

        sinks = SinkParticles(particles[[4, 7]], sink_radius=[1, 2] | units.AU)
        self.assertEqual(sinks.sink_radius, [1.0, 2.0] | units.AU)
        self.assertEqual(sinks.radius, 42.0 | units.RSun)
        self.assertEqual(sinks.mass, [5.0, 8.0] | units.MSun)
        self.assertEqual(sinks.position, [[4, 8, 12], [7, 14, 21]] | units.parsec)

        self.assertEqual(set(['key', 'mass', 'radius', 'x', 'y', 'z', 'sink_radius', 'vx', 'vy', 'vz', 'lx', 'ly', 'lz']),
            set(str(sinks).split("\n")[0].split()))
        self.assertEqual(set(['key', 'mass', 'radius', 'x', 'y', 'z']),
            set(str(particles).split("\n")[0].split()))

    def test4(self):
        print("Testing SinkParticles accrete")
        particles = Particles(10)
        particles.radius = 42.0 | units.RSun
        particles.mass = range(1, 11) | units.MSun
        particles.position = [[i, 2*i, 3*i] for i in range(10)] | units.parsec
        particles.velocity = [[i, 0, -i] for i in range(10)] | units.km/units.s
        particles.age = range(10) | units.Myr

        copy = particles.copy()

        sinks = SinkParticles(particles[[3, 7]], sink_radius=[4, 5] | units.parsec, looping_over=self.looping_over)
        self.assertEqual(sinks.sink_radius, [4.0, 5.0] | units.parsec)
        self.assertEqual(sinks.mass, [4.0, 8.0] | units.MSun)
        self.assertEqual(sinks.position, [[3, 6, 9], [7, 14, 21]] | units.parsec)

        accreted = sinks.accrete(particles)
        self.assertEqual(len(accreted), 4)  # 4 particles were accreted
        self.assertEqual(len(particles), 6)
        self.assertEqual(sinks.mass, [12.0, 24.0] | units.MSun)  # mass of sinks increased
        self.assertEqual(sinks.get_intersecting_subset_in(particles).mass,
            [12.0, 24.0] | units.MSun)  # original particles' masses match
        self.assertEqual(particles.total_mass(), copy.total_mass())  # total mass is conserved
        self.assertEqual(particles.center_of_mass(), copy.center_of_mass())  # center of mass is conserved
        self.assertEqual(particles.center_of_mass_velocity(), copy.center_of_mass_velocity())  # center of mass velocity is conserved
        self.assertEqual(particles.total_momentum(), copy.total_momentum())  # momentum is conserved
        self.assertEqual(particles.total_angular_momentum()+sinks.angular_momentum.sum(axis=0), copy.total_angular_momentum())  # angular_momentum is conserved

        sinks.sink_radius = [4.0, 8.0] | units.parsec
        accreted = sinks.accrete(particles)
        self.assertEqual(len(accreted), 2)  # another 2 particles were accreted
        self.assertEqual(len(particles), 4)
        self.assertEqual(sinks.mass, [12.0, 40.0] | units.MSun)  # mass of sinks increased
        self.assertEqual(sinks.get_intersecting_subset_in(particles).mass,
            [12.0, 40.0] | units.MSun)  # original particles' masses match
        self.assertEqual(particles.total_mass(), copy.total_mass())  # total mass is conserved
        self.assertEqual(particles.center_of_mass(), copy.center_of_mass())  # center of mass is conserved
        self.assertEqual(particles.center_of_mass_velocity(), copy.center_of_mass_velocity())  # center of mass velocity is conserved
        self.assertEqual(particles.total_momentum(), copy.total_momentum())  # momentum is conserved
        self.assertEqual(particles.total_angular_momentum()+sinks.angular_momentum.sum(axis=0), copy.total_angular_momentum())  # angular_momentum is conserved

    def test5(self):
        print("Testing SinkParticles accrete, one particle within two sinks' radii")
        particles = Particles(10)
        particles.radius = 42.0 | units.RSun
        particles.mass = range(1, 11) | units.MSun
        particles.position = [[i, 2*i, 3*i] for i in range(10)] | units.parsec
        particles.velocity = [[i, 0, -i] for i in range(10)] | units.km/units.s
        particles.age = range(10) | units.Myr
        copy = particles.copy()

        sinks = SinkParticles(particles[[3, 7]], sink_radius=[4, 12] | units.parsec, looping_over=self.looping_over)
        self.assertEqual(sinks.sink_radius, [4.0, 12.0] | units.parsec)
        self.assertEqual(sinks.mass, [4.0, 8.0] | units.MSun)
        self.assertEqual(sinks.position, [[3, 6, 9], [7, 14, 21]] | units.parsec)

        sinks.accrete(particles)
        self.assertEqual(len(particles), 4)  # 6 particles were accreted
        self.assertEqual(sinks.mass, [12.0, 40.0] | units.MSun)  # mass of sinks increased
        self.assertEqual(sinks.get_intersecting_subset_in(particles).mass,
            [12.0, 40.0] | units.MSun)  # original particles' masses match
        self.assertEqual(particles.total_mass(), copy.total_mass())  # total mass is conserved
        self.assertEqual(particles.center_of_mass(), copy.center_of_mass())  # center of mass is conserved
        self.assertEqual(particles.center_of_mass_velocity(), copy.center_of_mass_velocity())  # center of mass velocity is conserved
        self.assertEqual(particles.total_momentum(), copy.total_momentum())  # momentum is conserved
        self.assertEqual(particles.total_angular_momentum()+sinks.angular_momentum.sum(axis=0), copy.total_angular_momentum())  # angular_momentum is conserved


class TestSinkParticlesLoopingOverSources(TestSinkParticles):

    looping_over = "sources"


class TestNewSinkParticles(TestCase):

    looping_over = "sinks"

    def test1(self):
        print("Test the documentation for new_sink_particles")
        print(new_sink_particles.__doc__)

    def test2(self):
        print("Demonstrate new_sink_particles usage")
        cloud = Particles(100)
        cloud.mass = 1 | units.MSun
        cloud.position = [[0, 0, 0], [100, 100, 100], [200, 200, 200], [300, 300, 300]]*25 | units.parsec
        cloud.velocity = [[0, 0, 0], [1, 1, 1]]*50 | units.km / units.s
        unit_converter = ConvertBetweenGenericAndSiUnits(1 | units.m, 1 | units.kg, 1 | units.s)
        sph_code = Stub(unit_converter)
        sph_code.parameters.stopping_condition_maximum_density = 1 | units.kg / units.m**3
        sph_code.gas_particles.add_particles(cloud)
        density_limit_detection = sph_code.stopping_conditions.density_limit_detection
        density_limit_detection.enable()

        sph_code.evolve_model(1 | units.Myr)
        self.assertTrue(density_limit_detection.is_set())
        self.assertEqual(len(density_limit_detection.particles()), 3)
        self.assertEqual(density_limit_detection.particles().position,
            [[100, 100, 100], [200, 200, 200], [300, 300, 300]] | units.parsec)
        print(density_limit_detection.particles())

        clumps = density_limit_detection.particles().copy()
        sph_code.gas_particles.remove_particles(clumps)

        sinks = new_sink_particles(clumps, sink_radius=1 | units.parsec, looping_over=self.looping_over)
        self.assertEqual(sinks.sink_radius, 1.0 | units.parsec)
        self.assertEqual(sinks.mass, 1.0 | units.MSun)
        self.assertEqual(sinks.position,
            [[100, 100, 100], [200, 200, 200], [300, 300, 300]] | units.parsec)
        self.assertEqual(len(sph_code.gas_particles), 97)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass() + clumps.total_mass(), 100 | units.MSun, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), 97 | units.MSun, 10)

        sinks.accrete(sph_code.gas_particles)
        self.assertAlmostRelativeEqual(sinks.mass, [25, 25, 25] | units.MSun, 10)
        self.assertEqual(len(sph_code.gas_particles), 25)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass() + clumps.total_mass(), 100 | units.MSun, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), 25 | units.MSun, 10)

    def test3(self):
        print("Demonstrate new_sink_particles usage (using Gadget2)")
        UnitLength = 1.0 | units.kpc
        UnitMass = 1.0e10 | units.MSun
        UnitVelocity = 1.0 | units.km / units.s
        convert_nbody = nbody_system.nbody_to_si(UnitLength, UnitMass)
        converter = ConvertBetweenGenericAndSiUnits(UnitLength, UnitMass, UnitVelocity)
        number_gas_particles = 1000
        gas = new_evrard_gas_sphere(number_gas_particles, convert_nbody, do_scale=True, seed=12345)

        sph_code = Gadget2(converter)
        sph_code.initialize_code()
        sph_code.parameters.stopping_condition_maximum_density = 10 * UnitMass / UnitLength**3
        sph_code.gas_particles.add_particles(gas)
        self.assertIsOfOrder(max(sph_code.gas_particles.density), UnitMass / UnitLength**3)

        density_limit_detection = sph_code.stopping_conditions.density_limit_detection
        density_limit_detection.enable()

        sph_code.evolve_model(10.0 | units.Myr)
        self.assertTrue(density_limit_detection.is_set())
        self.assertTrue(sph_code.model_time < 10.0 | units.Myr)
        print("density_limit exceeded at t =", sph_code.model_time.as_quantity_in(units.Myr))
        self.assertEqual(len(density_limit_detection.particles()), 1)
        self.assertTrue(density_limit_detection.particles().density >
                10 * UnitMass / UnitLength**3)

        clumps = density_limit_detection.particles().copy()
        sph_code.gas_particles.remove_particles(clumps)
        clumps_in_code = sph_code.dm_particles.add_particles(clumps)

        sinks = new_sink_particles(clumps_in_code, looping_over=self.looping_over)
        self.assertEqual(sinks.sink_radius, clumps.radius)
        self.assertAlmostRelativeEqual(sinks.mass, UnitMass / number_gas_particles, 10)
        self.assertAlmostRelativeEqual(sinks.position, clumps.position, 10)
        self.assertEqual(len(sph_code.gas_particles), number_gas_particles - 1)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass(), UnitMass, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), UnitMass - sinks.total_mass(), 10)
        self.assertEqual(set(sinks.get_attribute_names_defined_in_store()) - set(["sink_radius", "lx", "ly", "lz"]),
            set(sph_code.particles.get_attribute_names_defined_in_store()))

        sinks.accrete(sph_code.gas_particles)
        self.assertAlmostRelativeEqual(sinks.mass, 3 * UnitMass / number_gas_particles, 10)
        self.assertEqual(len(sph_code.gas_particles), number_gas_particles - 3)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass(), UnitMass, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), UnitMass - sinks.total_mass(), 10)

        sinks.accrete(sph_code.particles)  # Nothing happens: gas already gone, and cannot accrete itself
        self.assertAlmostRelativeEqual(sinks.mass, 3 * UnitMass / number_gas_particles, 10)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass(), UnitMass, 10)

        steps = 0
        while True:
            sph_code.evolve_model(sph_code.model_time + (0.1 | units.Myr))
            sinks.sink_radius = 4 * clumps_in_code.radius
            sinks.accrete(sph_code.gas_particles)
            steps += 1
            if density_limit_detection.is_set():
                break

        self.assertEqual(len(sph_code.gas_particles), number_gas_particles - 7)
        self.assertAlmostRelativeEqual(sinks.mass, 7 * UnitMass / number_gas_particles, 10)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass(), UnitMass, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), UnitMass - sinks.total_mass(), 10)

        self.assertTrue(density_limit_detection.is_set())
        self.assertEqual(steps, 5)
        self.assertTrue(sph_code.model_time < 10.0 | units.Myr)
        print("density_limit exceeded at t =", sph_code.model_time.as_quantity_in(units.Myr))
        self.assertEqual(len(density_limit_detection.particles()), 5)
        self.assertTrue((density_limit_detection.particles().density >
                10 * UnitMass / UnitLength**3).all())

        clumps = density_limit_detection.particles().copy()
        sph_code.gas_particles.remove_particles(clumps)
        clumps_in_code = sph_code.dm_particles.add_particles(clumps)

        sinks.add_sinks(clumps_in_code, sink_radius=0.1 | units.kpc)
        self.assertEqual(sinks[1:].sink_radius, 0.1 | units.kpc)
        self.assertEqual(len(sph_code.gas_particles), number_gas_particles - 12)
        self.assertAlmostRelativeEqual(sinks.mass[1:], UnitMass / number_gas_particles, 10)
        self.assertAlmostRelativeEqual(sinks.position[1:], clumps.position, 10)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass(), UnitMass, 10)
        self.assertAlmostRelativeEqual(sph_code.gas_particles.total_mass(), UnitMass - sinks.total_mass(), 10)

        sinks.accrete(sph_code.gas_particles)
        self.assertEqual(len(sph_code.gas_particles), number_gas_particles - 66)
        self.assertAlmostRelativeEqual(sph_code.particles.total_mass().as_quantity_in(units.MSun), UnitMass, 10)
        self.assertAlmostRelativeEqual(sinks.mass, [7.0, 13.0, 15.0, 9.0, 11.0, 11.0] * UnitMass / number_gas_particles, 10)


class TestNewSinkParticlesLoopingOverSources(TestNewSinkParticles):

    looping_over = "sources"


class TestNonSphericalSinkParticles(TestCase):

    def test1(self):
        """ Test creating shapes and combining them """
        shapes = sink.Sphere(1 | units.RSun) | sink.Sphere(2 | units.RSun)
        self.assertEqual(len(shapes.sub_shapes), 2)
        shapes = sink.Sphere(1 | units.RSun) | sink.Sphere(2 | units.RSun) | sink.Spheroid([0.5, 0.5, 10] | units.RSun)
        self.assertEqual(len(shapes.sub_shapes), 3)

    def test2(self):
        """ Test Spheroid sink accretion """
        particles = self.create_particle_grid()

        spheroid = sink.Spheroid([5., 4., 1.] | units.RSun)
        sink_particles = Particles(1, mass=10. | units.MSun, radius=0. | units.RSun, position=[[1., 1., 1.]] | units.RSun)
        sinks = new_sink_particles(sink_particles, shapes=spheroid)

        accreted = sinks.accrete(particles)
        self.assertEqual(len(accreted), 161)
        self.assertEqual(len(particles), 1839)
        self.assertEqual(accreted.x.max(), 4 | units.RSun)
        self.assertEqual(accreted.y.max(), 3 | units.RSun)
        self.assertEqual(accreted.z.max(), 0.5 | units.RSun)

        spheroid.dimensions[2] = 3 | units.RSun
        accreted = sinks.accrete(particles)
        self.assertEqual(len(accreted), 324)
        self.assertEqual(len(particles), 1515)
        self.assertEqual(accreted.x.max(), 4 | units.RSun)
        self.assertEqual(accreted.y.max(), 3 | units.RSun)
        self.assertEqual(accreted.z.max(), 2.5 | units.RSun)
        self.assertIsSubvector([3, 2, 2.5] | units.RSun, particles.position)

    def test3(self):
        """ Test compound sink accretion """
        particles = self.create_particle_grid()

        shape = sink.Sphere(3. | units.RSun) | sink.Disc(*[5., 1.] | units.RSun)
        sink_particles = Particles(1, mass=10. | units.MSun, radius=0. | units.RSun, position=[[1., 1., 1.]] | units.RSun)
        sinks = new_sink_particles(sink_particles, shapes=shape)

        accreted = sinks.accrete(particles)
        self.assertEqual(len(accreted), 319)
        self.assertEqual(len(particles), 1681)
        self.assertEqual(accreted.x.max(), 4 | units.RSun)
        self.assertEqual(accreted.y.max(), 4 | units.RSun)
        self.assertEqual(accreted.z.max(), 2.5 | units.RSun)
        self.assertIsSubvector([2, 2, 1] | units.RSun, particles.position)

    def create_particle_grid(self):
        particles = Particles(2000)
        particles.radius = 1. | units.RSun
        particles.mass = 1. | units.MSun
        particles.velocity = [1, 0, -1] | units.km/units.s
        i = 0
        for x in range(-5, 5):
            for y in range(-5, 5):
                for z in numpy.arange(-5, 5, 0.5):
                    particles[i].position = [x, y, z] | units.RSun
                    i += 1
        return particles

    def assertIsSubvector(self, subvector, vectorlist):
        self._check_comparable(subvector, vectorlist)

        if not (subvector == vectorlist).all(axis=1).any():
            raise self.failureException("{} is not a subvector of given vectorlist".format(subvector))


class StubInterface(object):

    def __init__(self, **options):
        self.maximum_density = 1 | units.kg / units.m**3
        self._gas_particles = Particles()
        self._dm_particles = Particles()
        self._all_particles = ParticlesSuperset([self._gas_particles, self._dm_particles])

    def before_get_parameter(self):
        pass

    def before_set_parameter(self):
        pass

    def initialize_code(self):
        return 0

    synchronize_model = commit_particles = recommit_particles = commit_parameters = initialize_code

    def new_particle(self, mass, x, y, z, vx, vy, vz, *args):
        next_id = len(self._dm_particles)
        temp = Particles(len(mass))
        temp.mass = mass
        temp.x = x
        temp.y = y
        temp.z = z
        temp.vx = vx
        temp.vy = vy
        temp.vz = vz
        temp.id = range(next_id, next_id + len(mass))
        self._dm_particles.add_particles(temp)
        return [temp.id, temp.id]

    def new_gas_particle(self, mass, x, y, z, vx, vy, vz, *args):
        next_id = len(self._gas_particles) + 1000000
        temp = Particles(len(mass))
        temp.mass = mass
        temp.x = x
        temp.y = y
        temp.z = z
        temp.vx = vx
        temp.vy = vy
        temp.vz = vz
        temp.id = range(next_id, next_id + len(mass))
        self._gas_particles.add_particles(temp)
        return [temp.id, temp.id]

    def delete_particle(self, indices):
        for index in indices:
            for id, particle in zip(self._all_particles.id, self._all_particles):
                if id == index:
                    self._all_particles.remove_particle(particle)
        return 0

    def get_mass(self, indices):
        return [[mass for index in indices for id, mass in zip(self._all_particles.id,
            self._all_particles.mass) if index == id], [0]*len(indices)]

    def set_mass(self, indices, masses):
        for index, mass in zip(indices, masses):
            for id, particle in zip(self._all_particles.id, self._all_particles):
                if id == index:
                    particle.mass = mass
                    break
        return 0

    def get_position(self, indices):
        return [[x for index in indices for id, x in zip(self._all_particles.id, self._all_particles.x) if index == id],
            [y for index in indices for id, y in zip(self._all_particles.id, self._all_particles.y) if index == id],
            [z for index in indices for id, z in zip(self._all_particles.id, self._all_particles.z) if index == id],
            [0]*len(indices)]

    def get_velocity(self, indices):
        return [[vx for index in indices for id, vx in zip(self._all_particles.id, self._all_particles.vx) if index == id],
            [vy for index in indices for id, vy in zip(self._all_particles.id, self._all_particles.vy) if index == id],
            [vz for index in indices for id, vz in zip(self._all_particles.id, self._all_particles.vz) if index == id],
            [0]*len(indices)]

    def has_stopping_condition(self, type):
        return 1 if type == 6 else 0

    def get_stopping_condition_maximum_density_parameter(self):
        return self.maximum_density

    def set_stopping_condition_maximum_density_parameter(self, value):
        self.maximum_density = value

    is_stopping_condition_set = is_stopping_condition_enabled = has_stopping_condition

    def get_number_of_stopping_conditions_set(self):
        return 3

    def get_stopping_condition_info(self, sc_indices):
        return [6]*len(sc_indices), [1]*len(sc_indices)

    def get_stopping_condition_particle_index(self, sc_index, sc_sub_index):
        return list(range(len(self._gas_particles) + 1000000 - len(sc_index), len(self._gas_particles) + 1000000))

    def enable_stopping_condition(self, type):
        pass

    def evolve_model(self, time):
        return 0


class Stub(GravitationalDynamics):

    def __init__(self, unit_converter=None, **options):
        self.stopping_conditions = StoppingConditions(self)

        GravitationalDynamics.__init__(
            self,
            StubInterface(**options),
            unit_converter,
            **options
        )

    def define_parameters(self, object):
        object.add_method_parameter(
            "get_stopping_condition_maximum_density_parameter",
            "set_stopping_condition_maximum_density_parameter",
            "stopping_condition_maximum_density",
            "maximum density of a gas particle",
            default_value=-1.0 | generic_unit_system.density
        )

    def define_methods(self, object):
        GravitationalDynamics.define_methods(self, object)
        object.add_method("new_gas_particle",
            (nbody_system.mass, nbody_system.length, nbody_system.length, nbody_system.length,
                nbody_system.speed, nbody_system.speed, nbody_system.speed),
            (object.INDEX, object.ERROR_CODE))

    def define_particle_sets(self, object):
        object.define_super_set('particles', ['dm_particles', 'gas_particles'],
            index_to_default_set=0)

        object.define_set('dm_particles', 'index_of_the_particle')
        object.set_new('dm_particles', 'new_particle')
        object.set_delete('dm_particles', 'delete_particle')
        object.add_getter('dm_particles', 'get_mass', names=("mass",))
        object.add_setter('dm_particles', 'set_mass', names=("mass",))
        object.add_getter('dm_particles', 'get_position', names=("x", "y", "z"))
        object.add_getter('dm_particles', 'get_velocity', names=("vx", "vy", "vz"))

        object.define_set('gas_particles', 'index_of_the_particle')
        object.set_new('gas_particles', 'new_gas_particle')
        object.set_delete('gas_particles', 'delete_particle')
        object.add_getter('gas_particles', 'get_mass', names=("mass",))
        object.add_getter('gas_particles', 'get_position', names=("x", "y", "z"))
        object.add_getter('gas_particles', 'get_velocity', names=("vx", "vy", "vz"))

        object.add_query('particles', 'get_stopping_condition_particle_index')



import os
import os.path
import numpy

from amuse.units import units
from amuse.datamodel import Particles
from amuse.support.exceptions import AmuseException
from amuse.test.amusetest import TestCase, get_path_to_results

from amuse.community.mesa.interface import MESA
from amuse.community.evtwin.interface import EVtwin
from amuse.community.sse.interface import SSE

from amuse.couple.parallel_stellar_evolution import ParallelStellarEvolution


default_options = dict(must_run_threaded=False)  # Always thread-safe
# ~default_options = dict() # Really really parallel


class TestParallelStellarEvolution(TestCase):

    code_factory = SSE

    def test1(self):
        print("Testing ParallelStellarEvolution initialization")
        instance = ParallelStellarEvolution(self.code_factory, number_of_workers=3, **default_options)
        instance.initialize_code()
        instance.cleanup_code()
        instance.stop()

    def test2(self):
        print("Testing ParallelStellarEvolution particles")
        instance = ParallelStellarEvolution(self.code_factory, number_of_workers=2, **default_options)
        instance.initialize_code()
        instance.commit_parameters()

        particles = Particles(5)
        particles.mass = range(1, 1+len(particles)) | units.MSun
        incode = instance.particles.add_particles(particles)
        instance.commit_particles()
        self.assertAlmostEqual(incode.mass, range(1, 1+len(particles)) | units.MSun)
        print("Note that the order of instance.particles is different from the", end=' ')
        print("original particle order, since particles are distributed over 2 processes")
        self.assertAlmostEqual(instance.particles.mass, [1, 3, 5, 2, 4] | units.MSun)

        instance.stop()

    def slowtest3(self):
        print("Testing ParallelStellarEvolution evolve_model")
        particles = Particles(4)
        particles.mass = range(1, 1+len(particles)) | units.MSun

        serial = MESA()
        inserial = serial.particles.add_particles(particles)
        self.assertAlmostEqual(inserial.mass, range(1, 1+len(particles)) | units.MSun)
        serial.evolve_model(0.2 | units.Myr)

        parallel = ParallelStellarEvolution(MESA, number_of_workers=3, **default_options)
        inparallel = parallel.particles.add_particles(particles)
        self.assertAlmostEqual(inparallel.mass, range(1, 1+len(particles)) | units.MSun)
        parallel.evolve_model(0.2 | units.Myr)
        self.assertEqual(parallel.model_time, 0.2 | units.Myr)
        self.assertTrue(numpy.all(inparallel.age >= (0.2 | units.Myr)))
        self.assertTrue(numpy.all(inparallel.age - inparallel.time_step <= (0.2 | units.Myr)))

        self.assertEqual(inserial.luminosity, inparallel.luminosity)
        self.assertEqual(inserial.time_step, inparallel.time_step)
        self.assertEqual(inserial.temperature, inparallel.temperature)
        serial.stop()
        parallel.stop()

    def test4(self):
        print("Testing ParallelStellarEvolution parameters")
        parallel = ParallelStellarEvolution(self.code_factory, number_of_workers=3, **default_options)
        parallel.parameters.metallicity = 0.01
        self.assertEqual(parallel.parameters.metallicity, 0.01)
        for code in parallel.code_instances:
            self.assertEqual(code.parameters.metallicity, 0.01)
        parallel.stop()

    def test5(self):
        print("Testing ParallelStellarEvolution individual options")
        base_name = os.path.join(get_path_to_results(), "parallel_stellar_evolution_out_")
        for filename in [base_name+str(i) for i in range(3)]:
            if os.path.exists(filename):
                os.remove(filename)

        parallel = ParallelStellarEvolution(self.code_factory, number_of_workers=3,
            individual_options=[dict(redirect_file=base_name+str(i)) for i in range(3)], redirection="file", **default_options)

        for filename in [base_name+str(i) for i in range(3)]:
            self.assertTrue(os.path.exists(filename))

        parallel.stop()

    def xtest6(self):
        print("Testing ParallelStellarEvolution exception handling")
        self.code_factory = EVtwin
        if self.code_factory == MESA:
            expected_message = ("Error when calling 'evolve_for' of a 'MESA', errorcode is -12, error is 'Evolve terminated: Maximum age reached.'")
        elif self.code_factory == EVtwin:
            expected_message = ("Error when calling 'evolve_for' of a 'EVtwin', errorcode is 5, error is 'PRINTB -- age greater than limit'")
        else:
            self.skip("Skipping test: {0} has no parameter max_age_stop_condition".format(self.code_factory))

        parallel = ParallelStellarEvolution(self.code_factory, number_of_workers=2, **default_options)
        parallel.parameters.max_age_stop_condition = 0.1 | units.Myr

# ~        parallel.particles.add_particles(Particles(2, mass=[1,2]|units.MSun)) # Test speed-up:
        parallel.particles.add_particles(Particles(1, mass=1 | units.MSun))

        self.assertRaises(AmuseException, parallel.evolve_model, 1.0 | units.Myr,
            expected_message=expected_message)
        self.assertTrue((parallel.particles.age >= 0.1 | units.Myr).all())
        self.assertTrue((parallel.particles.age-parallel.particles.time_step <= 0.1 | units.Myr).all())
        parallel.stop()



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI

from amuse import datamodel
from amuse.rfi.core import PythonCodeInterface, legacy_function, LegacyFunctionSpecification

from amuse.support import exceptions

import numpy
from mpi4py import MPI


class ForTestingInterface(PythonCodeInterface):

    def __init__(self, **options):
        PythonCodeInterface.__init__(self, implementation_factory=ForTestingImplementation, **options)

    @legacy_function
    def get_range():
        function = LegacyFunctionSpecification()
        function.addParameter('imin', dtype='int32', direction=function.OUT)
        function.addParameter('imax', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def get_x():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float64', direction=function.OUT)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def get_dens():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('dens', dtype='float64', direction=function.OUT)
        function.addParameter('N', dtype='int32', direction=function.LENGTH)

        function.result_type = 'int32'
        function.must_handle_array = True
        return function


class ForTestingImplementation(object):

    def __init__(self):
        self.comm = MPI.COMM_WORLD
        self.myrank = self.comm.Get_rank()
        self.N = self.comm.Get_size()
        self.Ngrid = 3*4*5
        n = self.Ngrid//self.N
        x = (numpy.arange(n)+self.myrank*n)/(1.*self.Ngrid)
        self.local_imin = self.myrank*n
        self.local_imax = (self.myrank+1)*n-1
        self.dens = x**2

    def get_range(self, imin, imax):
        imin.value = 0
        imax.value = self.Ngrid-1
        return 0

    def get_x(self, index, x):
        x.value = index/(1.*self.Ngrid)
        return 0

    def get_dens(self, index, dens, N):
        a = (index >= self.local_imin)*(index <= self.local_imax)
        _dens = numpy.zeros(N)
        _dens[a] = self.dens[index[a]-self.local_imin]
        dens.value = numpy.zeros(N)
        _dens = self.comm.Reduce(_dens, dens.value, MPI.SUM, root=0)
        return 0


class ForTesting(InCodeComponentImplementation):

    def __init__(self, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(**options), **options)

    def define_grids(self, object):
        object.define_grid('grid', axes_names=['x'], grid_class=datamodel.CartesianGrid)
        object.set_grid_range('grid', 'get_range')
        object.add_getter('grid', 'get_dens', names=('dens',))
        object.add_getter('grid', 'get_x', names=('x',))


class TestInterface(TestWithMPI):

    def ForTesting(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        return ForTesting(**options)

    def ForTestingInterface(self, **options):
        options["worker_dir"] = self.get_path_to_results()
        return ForTestingInterface(**options)

    def test1(self):
        interface = self.ForTesting(redirection="none", number_of_workers=1)
        x = interface.grid.x
        dens = interface.grid.dens
        self.assertEqual(x, numpy.arange(60)/60.)
        self.assertEqual(dens, x**2)
        interface.stop()

    def test2(self):
        for n in [3, 5, 6]:
            interface = self.ForTesting(redirection="none", number_of_workers=n)
            x = interface.grid.x
            dens = interface.grid.dens
            self.assertEqual(x, numpy.arange(60)/60.)
            self.assertEqual(dens, x**2)
            interface.stop()



from amuse.test import amusetest

from amuse.datamodel.grids import *

import numpy


class TestGridAttributes(amusetest.TestCase):

    def test1(self):
        grid = new_cartesian_grid((10, 20), 1.)
        self.assertEqual(grid.cellsize(), [1., 1.])
        self.assertEqual(grid.get_minimum_index(), (0, 0))
        self.assertEqual(grid.get_maximum_index(), (9, 19))
        self.assertEqual(grid.get_minimum_position(), [0., 0.])
        self.assertEqual(grid.get_maximum_position(), [10., 20.])
        self.assertEqual(grid.get_volume(), 10*20.)
        self.assertEqual(grid.contains(numpy.asarray([[1, 1], [-10, -10]])), [True, False])
        points = grid.points()
        x = points[:, :, 0]
        y = points[:, :, 1]
        self.assertEqual([x, y], numpy.indices((11, 21)))

        grid2 = new_cartesian_grid((10, 20), 1., offset=(5, 5))
        self.assertTrue(grid.overlaps(grid2))
        grid3 = new_cartesian_grid((10, 20), 1., offset=(15, 25))
        self.assertFalse(grid.overlaps(grid3))
        overlap = grid.get_overlap_with(grid2)
        self.assertEqual(overlap.shape, (5, 15))

    def test2(self):
        grid = new_regular_grid((10, 20), (10, 10))
        self.assertEqual(grid.cellsize(), [1., 0.5])
        self.assertEqual(grid.get_minimum_index(), (0, 0))
        self.assertEqual(grid.get_maximum_index(), (9, 19))
        self.assertEqual(grid.get_minimum_position(), [0., 0.])
        self.assertEqual(grid.get_maximum_position(), [10., 10.])
        self.assertEqual(grid.get_volume(), 10*10.)
        self.assertEqual(grid.contains(numpy.asarray([[1, 1], [-10, -10]])), [True, False])
        points = grid.points()
        x = points[:, :, 0]
        y = points[:, :, 1]
        xp, yp = numpy.indices((11, 21))
        self.assertEqual(x, xp)
        self.assertEqual(y, yp*0.5)

        grid2 = new_cartesian_grid((10, 20), 1., offset=(5, 5))
        self.assertTrue(grid.overlaps(grid2))
        grid3 = new_cartesian_grid((10, 20), 1., offset=(15, 25))
        self.assertFalse(grid.overlaps(grid3))
        overlap = grid.get_overlap_with(grid2)
        self.assertEqual(overlap.shape, (5, 10))

    def test3(self):
        x = numpy.arange(11)/10.
        y = numpy.arange(21)/20.
        grid = new_rectilinear_grid((10, 20), (x, y**2))
        self.assertRaises(Exception, grid.cellsize,
          expected_message="a RectilinearGrid does not have a constant cellsize, use the cellsizes method instead")

    def xtest4(self):
        grid = new_structured_grid((10, 20), (10, 10))
        self.assertEqual(grid.cellsize(), [1., 0.5])

    def xtest5(self):
        grid = new_unstructured_grid((10, 20), (10, 10))
        self.assertEqual(grid.cellsize(), [1., 0.5])

    def test6(self):
        grid = new_regular_grid((10, 10), (10, 10))
        self.assertEqual(grid.get_index((6.2, 3.7)), [6, 3])
        self.assertEqual(grid.get_index(x=[6.2], y=[3.7]), [6, 3])
        self.assertEqual(grid.get_index(y=[6.2], x=[3.7]), [3, 6])
        self.assertEqual(grid.get_index(y=6.2, x=3.7)[0], 3)
        self.assertEqual(grid.get_index(y=6.2, x=3.7)[1], 6)

    def test7(self):
        grid = new_regular_grid((10, 10), [20, 10] | units.m, axes_names="ab")
        self.assertEqual(grid.get_index([16.2, 3.7] | units.m), [8, 3])
        self.assertEqual(grid.get_index(a=16.2 | units.m, b=3.7 | units.m), [8, 3])
        self.assertEqual(grid.get_index(a=[16.2, 4.5] | units.m, b=[3.7, 4.2] | units.m), [[8, 3], [2, 4]])



import numpy
from amuse.test.amusetest import TestWithMPI

from amuse.community.fi.interface import Fi
from amuse.community.hermite.interface import Hermite
from amuse.community.phigrape.interface import PhiGRAPE
from amuse.community.bhtree.interface import BHTree
from amuse.ext.bridge import bridge
from amuse.units import nbody_system
from amuse.units import units
from amuse.ic.kingmodel import new_king_model


def sys_from_parts(base_class, parts, converter, eps=None):
    interface = base_class(converter)
    interface.initialize_code()
    if eps is not None:
        interface.parameters.epsilon_squared = eps**2
    interface.particles.add_particles(parts)
    return interface


class TestBridge(TestWithMPI):

    def test1(self):
        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        eps = 2.e-4 | nbody_system.length

        test_class = PhiGRAPE
        number_of_particles = 50
        stars = new_king_model(number_of_particles, W0=7, convert_nbody=convert)
        stars.radius = 0.0 | units.RSun

        cluster = test_class(convert)
        cluster.parameters.epsilon_squared = eps**2
        cluster.particles.add_particles(stars)
        cluster.synchronize_model()

        Ep1 = convert.to_nbody(cluster.potential_energy).number
        Ek1 = convert.to_nbody(cluster.kinetic_energy).number

        parts = cluster.particles.copy()
        parts1 = parts.select_array(lambda x: (x > 0 | units.m), ['x'])
        parts2 = parts.select_array(lambda x: (x < 0 | units.m), ['x'])
        cluster1 = sys_from_parts(test_class, parts1, convert, eps)
        cluster2 = sys_from_parts(test_class, parts2, convert, eps)

        cluster1.synchronize_model()
        cluster2.synchronize_model()

        bridgesys = bridge()
        bridgesys.add_system(cluster1, (cluster2,))
        bridgesys.add_system(cluster2, (cluster1,))

        Ep2 = convert.to_nbody(bridgesys.potential_energy).number
        Ek2 = convert.to_nbody(bridgesys.kinetic_energy).number
        self.assertAlmostEqual(Ek1, Ek2, 12)
        self.assertAlmostEqual(Ep1, Ep2, 12)



from amuse.test.amusetest import TestWithMPI
import os
import sys
import numpy
import math

from amuse.community.smalln.interface import SmallNInterface, SmallN

from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse import io
from amuse.datamodel import trees
from amuse.ic import plummer
try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestSmallNInterface(TestWithMPI):

    def test0(self):
        instance = SmallNInterface()
        instance.stop()

    def test1(self):
        instance = SmallNInterface()
        instance.initialize_code()

        res1 = instance.new_particle(mass=11.0, radius=2.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        res2 = instance.new_particle(mass=21.0, radius=5.0, x=10.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        self.assertEqual(1, res1['index_of_the_particle'])
        self.assertEqual(2, res2['index_of_the_particle'])

        retrieved_state1 = instance.get_state(1)
        retrieved_state2 = instance.get_state(2)

        self.assertEqual(11.0,  retrieved_state1['mass'])
        self.assertEqual(21.0,  retrieved_state2['mass'])
        self.assertEqual(0.0,  retrieved_state1['x'])
        self.assertEqual(10.0,  retrieved_state2['x'])
        self.assertEqual(2.0,  retrieved_state1['radius'])
        self.assertEqual(5.0,  retrieved_state2['radius'])

        instance.cleanup_code()
        instance.stop()

    def test2(self):
        instance = SmallNInterface()
        instance.initialize_code()
        self.skip("index of the next particle not implemented correctly yet")
        for i in [0, 1, 2]:
            temp_particle = instance.new_particle(mass=i, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
            self.assertEqual(i+1, temp_particle['index_of_the_particle'])

        instance.delete_particle(2)

        self.assertEqual(2, instance.get_number_of_particles()['number_of_particles'])

        self.assertEqual(1, instance.get_index_of_first_particle()['index_of_the_particle'])

        self.assertEqual(2, instance.get_index_of_next_particle(1)['index_of_the_next_particle'])
        self.assertEqual(0, instance.get_index_of_next_particle(1)['__result'])
        self.assertEqual(-1, instance.get_index_of_next_particle(3)['__result'])
        self.assertEqual(1, instance.get_index_of_next_particle(2)['__result'])
        instance.cleanup_code()
        instance.stop()

    def test5(self):
        smalln = SmallNInterface()
        smalln.initialize_code()

        smalln.new_particle([10, 20], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        retrieved_state = smalln.get_state(1)

        self.assertEqual(10.0,  retrieved_state['mass'])
        self.assertEqual(1, retrieved_state['radius'])

        retrieved_state = smalln.get_state([1, 2])
        self.assertEqual(20.0,  retrieved_state['mass'][1])
        self.assertEqual(smalln.get_number_of_particles()['number_of_particles'], 2)
        smalln.cleanup_code()
        smalln.stop()

    def test6(self):
        smalln = SmallNInterface()
        smalln.initialize_code()

        smalln.new_particle([10, 10], [-1, 1], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [1, 1])
        retrieved_state = smalln.get_state(1)

        self.assertFalse(hasattr(smalln, 'get_potential_at_point'))
        # retr = smalln.get_potential_at_point(0.01, 0, 0, 0)
        # self.assertEqual(retr['__result'], -1)
        smalln.cleanup_code()
        smalln.stop()

    def xtest7(self):
        instance = SmallNInterface()
        instance.initialize_code()
        instance.set_eps2(0.1 * 0.1)
        instance.commit_parameters()
        id1, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        id2, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=2.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        potential, errorcode = instance.get_potential(id1)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential,  -10.0 / numpy.sqrt(2.0**2 + 0.1**2), 8)

        potential, errorcode = instance.get_potential(id2)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential,  -10.0 / numpy.sqrt(2.0**2 + 0.1**2), 8)

        total_potential, errorcode = instance.get_potential_energy()
        potentials, errorcode = instance.get_potential([id1, id2])
        instance.cleanup_code()
        instance.stop()

        self.assertAlmostRelativeEquals(total_potential, numpy.sum(potentials * [10.0, 10.0]) / 2.0)

    def xtest8(self):
        instance = SmallNInterface()
        instance.initialize_code()
        instance.commit_parameters()
        id1, errorcode = instance.new_particle(mass=10.0, radius=1.0, x=0.0, y=0.0, z=0.0, vx=0.0, vy=0.0, vz=0.0)
        id2, errorcode = instance.new_particle(mass=1.0, radius=1.0, x=2.0, y=0.0, z=0.0, vx=10.0, vy=0.0, vz=0.0)

        instance.commit_particles()
        potential, errorcode = instance.get_potential(id1)
        self.assertEqual(errorcode, 0)
        self.assertAlmostRelativeEquals(potential,  -1.0 / numpy.sqrt(2.0**2), 8)
        total_potential, errorcode = instance.get_potential_energy()
        potentials, errorcode = instance.get_potential([id1, id2])
        instance.cleanup_code()
        instance.stop()

        self.assertAlmostRelativeEquals(total_potential, numpy.sum(potentials * [10.0, 1.0]) / 2.0)

    def test9(self):
        print("Test SmallNInterface evolve_model")
        instance = SmallNInterface()
        self.assertEqual(0, instance.initialize_code())
        self.assertEqual(0, instance.set_eta(0.001))
        self.assertEqual(0, instance.commit_parameters())

        # Set up an equal-mass binary on a circular orbit:
        self.assertEqual([1, 0], list(instance.new_particle(0.5,  0.5, 0, 0,  0, 0.5, 0, 0.01).values()))
        self.assertEqual([2, 0], list(instance.new_particle(0.5,  -0.5, 0, 0,  0, -0.5, 0, 0.01).values()))
        self.assertEqual(0, instance.commit_particles())

        self.assertEqual(0, instance.evolve_model(math.pi))
        for result, expected in zip(instance.get_position(1).values(), [-0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)
        for result, expected in zip(instance.get_position(2).values(), [0.5, 0.0, 0.0, 0]):
            self.assertAlmostEqual(result, expected, 3)

        self.assertEqual(0, instance.evolve_model(2 * math.pi))
        # print instance.get_time()
        # print instance.get_position(1), instance.get_velocity(1)
        # print instance.get_position(2)
        # for result, expected in zip(instance.get_position(1).values(), [0.5, 0.0, 0.0, 0]):
        #    self.assertAlmostEquals(result, expected, 3)
        # for result, expected in zip(instance.get_position(2).values(), [-0.5, 0.0, 0.0, 0]):
        #    self.assertAlmostEquals(result, expected, 3)

        self.assertEqual(0, instance.cleanup_code())
        instance.cleanup_code()
        instance.stop()


class TestSmallN(TestWithMPI):
    def new_system_of_sun_and_earth(self):
        stars = datamodel.Stars(2)
        sun = stars[0]
        sun.mass = units.MSun(1.0)
        sun.position = units.m(numpy.array((0.0, 0.0, 0.0)))
        sun.velocity = units.ms(numpy.array((0.0, 0.0, 0.0)))
        sun.radius = units.RSun(1.0)

        earth = stars[1]
        earth.mass = units.kg(5.9736e24)
        earth.radius = units.km(6371)
        earth.position = units.km(numpy.array((149.5e6, 0.0, 0.0)))
        earth.velocity = units.ms(numpy.array((0.0, 29800, 0.0)))

        return stars

    def test1(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        smalln = SmallN(convert_nbody)
        smalln.initialize_code()
        smalln.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]

        smalln.particles.add_particles(stars)

        smalln.evolve_model(365.0 | units.day)
        smalln.particles.copy_values_of_all_attributes_to(stars)

        position_at_start = earth.position.value_in(units.AU)[0]
        position_after_full_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(position_at_start, position_after_full_rotation, 6)

        smalln.evolve_model(365.0 + (365.0 / 2) | units.day)

        smalln.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[0]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 2)

        smalln.evolve_model(365.0 + (365.0 / 2) + (365.0 / 4) | units.day)

        smalln.particles.copy_values_of_all_attributes_to(stars)
        position_after_half_a_rotation = earth.position.value_in(units.AU)[1]
        self.assertAlmostEqual(-position_at_start, position_after_half_a_rotation, 3)

        smalln.cleanup_code()

        smalln.stop()

    def test2(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = SmallN(convert_nbody)
        instance.initialize_code()
        instance.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()
        earth = stars[1]
        instance.particles.add_particles(stars)

        for x in range(1, 2000, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        if HAS_MATPLOTLIB:
            figure = pyplot.figure()
            plot = figure.add_subplot(1, 1, 1)

            x_points = earth.get_timeline_of_attribute("x")
            y_points = earth.get_timeline_of_attribute("y")

            x_points_in_AU = [t_x[1].value_in(units.AU) for t_x in x_points]
            y_points_in_AU = [t_x1[1].value_in(units.AU) for t_x1 in y_points]

            plot.scatter(x_points_in_AU, y_points_in_AU, color="b", marker='o')

            plot.set_xlim(-1.5, 1.5)
            plot.set_ylim(-1.5, 1.5)

            test_results_path = self.get_path_to_results()
            output_file = os.path.join(test_results_path, "smalln-earth-sun2.svg")
            figure.savefig(output_file)

        instance.cleanup_code()
        instance.stop()

    def test3(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        instance = SmallN(convert_nbody)
        instance.initialize_code()
        instance.dt_dia = 5000

        stars = datamodel.Stars(2)
        star1 = stars[0]
        star2 = stars[1]

        star1.mass = units.MSun(1.0)
        star1.position = units.AU(numpy.array((-1.0, 0.0, 0.0)))
        star1.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star1.radius = units.RSun(1.0)

        star2.mass = units.MSun(1.0)
        star2.position = units.AU(numpy.array((1.0, 0.0, 0.0)))
        star2.velocity = units.AUd(numpy.array((0.0, 0.0, 0.0)))
        star2.radius = units.RSun(100.0)

        instance.particles.add_particles(stars)

        for x in range(1, 2000, 10):
            instance.evolve_model(x | units.day)
            instance.particles.copy_values_of_all_attributes_to(stars)
            stars.savepoint()

        instance.stop()

    def test4(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = SmallN(convert_nbody)
        instance.initialize_code()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 2)

        instance.particles.mass = [17.0, 33.0] | units.kg

        self.assertEqual(instance.get_mass(1), 17.0 | units.kg)
        self.assertEqual(instance.get_mass(2), 33.0 | units.kg)
        instance.stop()

    def test5(self):
        convert_nbody = nbody_system.nbody_to_si(5.0 | units.kg, 10.0 | units.m)

        instance = SmallN(convert_nbody)
        instance.initialize_code()

        particles = datamodel.Particles(2)
        self.assertEqual(len(instance.particles), 0)

        particles.mass = [15.0, 30.0] | units.kg
        particles.radius = [10.0, 20.0] | units.m
        particles.position = [[10.0, 20.0, 30.0], [20.0, 40.0, 60.0]] | units.m
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | units.m / units.s

        instance.particles.add_particles(particles)
        self.assertEqual(len(instance.particles), 2)
        instance.set_state(1, 16 | units.kg,
                           20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                           1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms,
                           20.0 | units.m)

        curr_state = instance.get_state(1)
        for expected, actual in zip([16 | units.kg,
                           20.0 | units.m, 40.0 | units.m, 60.0 | units.m,
                           1.0 | units.ms, 1.0 | units.ms, 1.0 | units.ms,
                           20.0 | units.m], curr_state):
            self.assertAlmostRelativeEquals(expected, actual)
        instance.stop()

        self.assertEqual(curr_state[0], 16 | units.kg, 8)

    def test6(self):
        print("Test6: Testing SmallN parameters")
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.yr, 1.0 | units.AU)
        instance = SmallN(convert_nbody)
        instance.initialize_code()

        value = instance.get_eta()
        self.assertEqual(0.14, value)
        self.assertAlmostEqual(0.14, instance.parameters.timestep_parameter)
        for x in [0.001, 0.01, 0.1]:
            instance.parameters.timestep_parameter = x
            self.assertAlmostEqual(x, instance.parameters.timestep_parameter)

        value = instance.get_time()
        self.assertEqual(0 | units.yr, value)

        value = instance.get_gamma()
        self.assertEqual(1e-6, value)
        self.assertAlmostEqual(1e-6, instance.parameters.unperturbed_threshold)
        for x in [0.001, 0.01, 0.1]:
            instance.parameters.unperturbed_threshold = x
            self.assertAlmostEqual(x, instance.parameters.unperturbed_threshold)
        instance.stop()

    def test15(self):
        particles = datamodel.Particles(2)
        particles.x = [0.0, 10.0] | nbody_system.length
        particles.y = 0.0 | nbody_system.length
        particles.z = 0.0 | nbody_system.length
        particles.vx = 0.0 | nbody_system.speed
        particles.vy = 0.0 | nbody_system.speed
        particles.vz = 0.0 | nbody_system.speed
        particles.mass = 1.0 | nbody_system.mass

        instance = SmallN()
        instance.particles.add_particles(particles)
        instance.commit_particles()
        self.assertEqual(instance.particles[0].radius, 0.0 | nbody_system.length)
        p = datamodel.Particle(
            x=1.0 | nbody_system.length,
            y=2.0 | nbody_system.length,
            z=3.0 | nbody_system.length,
            vx=1.0 | nbody_system.speed,
            vy=2.0 | nbody_system.speed,
            vz=3.0 | nbody_system.speed,
            mass=1.0 | nbody_system.mass,
            radius=4.0 | nbody_system.length,
        )
        instance.particles.add_particle(p)
        self.assertEqual(instance.particles[0].radius, 0.0 | nbody_system.length)
        self.assertEqual(instance.particles[1].radius, 0.0 | nbody_system.length)
        self.assertEqual(instance.particles[2].radius, 4.0 | nbody_system.length)

        instance.stop()

    def assertEarthAndMoonWasDetectedAsBinary(self, set, stars):
        alltrees = trees.BinaryTreesOnAParticleSet(set, 'child1', 'child2')
        roots = list(alltrees.iter_roots())
        self.assertEqual(len(roots), 1)
        root = roots[0].particle
        earth_and_moon = None
        if not root.child1.child1 is None:
            earth_and_moon = root.child1
        else:
            earth_and_moon = root.child2

        self.assertAlmostRelativeEquals(earth_and_moon.mass, stars[1].mass + stars[2].mass)

    def test16(self):
        convert_nbody = nbody_system.nbody_to_si(1.0 | units.MSun, 149.5e6 | units.km)

        smalln = SmallN(convert_nbody)
        smalln.initialize_code()
        smalln.dt_dia = 5000

        stars = self.new_system_of_sun_and_earth()

        moon = datamodel.Particle()
        moon.mass = units.kg(7.3477e22)
        moon.radius = units.km(1737.10)
        moon.position = units.km(numpy.array((149.5e6 + 384.399, 0.0, 0.0)))
        moon.velocity = units.ms(numpy.array((0.0, 29800 + 1022, 0.0)))

        stars.add_particle(moon)

        earth = stars[1]

        smalln.particles.add_particles(stars)

        smalln.evolve_model(365.0 | units.day)
        smalln.update_particle_tree()
        smalln.update_particle_set()

        self.assertEqual(len(smalln.particles), 5)

        self.assertEarthAndMoonWasDetectedAsBinary(smalln.particles, stars)

        inmemory = smalln.particles.copy()
        self.assertEarthAndMoonWasDetectedAsBinary(inmemory, stars)

        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "newsmalln-test16.hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        io.write_set_to_file(smalln.particles, output_file, "hdf5")
        fromfile = io.read_set_from_file(output_file, "hdf5")
        self.assertEarthAndMoonWasDetectedAsBinary(fromfile, stars)
        smalln.stop()

    def test17(self):

        particles = datamodel.Particles(keys=[1, 2, 3, 4, 5, 6, 7])
        particles.mass = 0.001 | nbody_system.mass
        particles.radius = 0.1 | nbody_system.length
        particles.x = [
            -100.5, -99.5,
              -0.5, 0.5,
              99.5, 100.5,
             120.0
        ] | nbody_system.length
        particles.y = 0 | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [
            [2, 0, 0], [-2, 0, 0],
            [2, 0, 0], [-2, 0, 0],
            [2, 0, 0], [-2, 0, 0],
            [-4, 0, 0]
        ] | nbody_system.speed

        instance = SmallN()
        instance.particles.add_particles(particles)

        collisions = instance.stopping_conditions.collision_detection
        collisions.enable()

        instance.evolve_model(1.0 | nbody_system.time)

        self.assertTrue(collisions.is_set())
        self.assertTrue(instance.model_time < 0.5 | nbody_system.time)

        self.assertEqual(len(collisions.particles(0)), 3)
        self.assertEqual(len(collisions.particles(1)), 3)
        self.assertEqual(len(particles - collisions.particles(0) - collisions.particles(1)), 1)
        self.assertEqual(abs(collisions.particles(0).x - collisions.particles(1).x) <=
                (collisions.particles(0).radius + collisions.particles(1).radius),
                [True, True, True])
        instance.stop()

    def test18(self):

        particles = datamodel.Particles(keys=[1, 2])
        particles.mass = 1 | nbody_system.mass
        particles.radius = 0.1 | nbody_system.length
        particles.x = [1, -1] | nbody_system.length
        particles.y = [1, -1] | nbody_system.length
        particles.z = 0 | nbody_system.length
        particles.velocity = [[-1, 0, 0], [1, 0, 0]] | nbody_system.speed

        instance = SmallN()
        instance.particles.add_particles(particles)

        stopping_condition = instance.stopping_conditions.interaction_over_detection
        stopping_condition.enable()

        instance.evolve_model(10.0 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())
        self.assertTrue(instance.model_time < 11.0 | nbody_system.time)
        instance.stop()

    def test19(self):

        particles = datamodel.Particles(keys=[1, 2, 3])
        particles.mass = 1 | nbody_system.mass
        particles.radius = 0.1 | nbody_system.length
        particles.x = [1, -1, 0] | nbody_system.length
        particles.y = [1, -1, 0] | nbody_system.length
        particles.z = [0, 0, 1] | nbody_system.length
        particles.velocity = [[-1, 0, 0], [1, 0, 0], [0, 0, -10]] | nbody_system.speed

        instance = SmallN()
        instance.particles.add_particles(particles)

        stopping_condition = instance.stopping_conditions.interaction_over_detection
        stopping_condition.enable()

        instance.evolve_model(10.0 | nbody_system.time)
        self.assertTrue(stopping_condition.is_set())

        self.assertTrue(instance.model_time < 10.0 | nbody_system.time)
        instance.stop()

    def test20(self):
        p = datamodel.Particles(3)

        p[0].mass = 6.667e-01 | nbody_system.mass
        p[0].radius = 4.000e-03 | nbody_system.length
        p[0].x = -1.309e+01 | nbody_system.length
        p[0].y = 1.940e+01 | nbody_system.length
        p[0].z = -1.163e+01 | nbody_system.length
        p[0].vx = 2.366e-01 | nbody_system.speed
        p[0].vy = -3.813e-01 | nbody_system.speed
        p[0].vz = 2.486e-01 | nbody_system.speed

        p[1].mass = 3.333e-01 | nbody_system.mass
        p[1].radius = 1.000e-03 | nbody_system.length
        p[1].x = -1.506e+01 | nbody_system.length
        p[1].y = 1.937e+01 | nbody_system.length
        p[1].z = -1.163e+01 | nbody_system.length
        p[1].vx = 3.483e-01 | nbody_system.speed
        p[1].vy = -4.513e-01 | nbody_system.speed
        p[1].vz = 2.486e-01 | nbody_system.speed

        p[2].mass = 5.000e-01 | nbody_system.mass
        p[2].radius = 2.000e-03 | nbody_system.length
        p[2].x = 2.749e+01 | nbody_system.length
        p[2].y = -3.877e+01 | nbody_system.length
        p[2].z = 2.325e+01 | nbody_system.length
        p[2].vx = -5.476e-01 | nbody_system.speed
        p[2].vy = 8.092e-01 | nbody_system.speed
        p[2].vz = -4.972e-01 | nbody_system.speed

        instance = SmallN()
        instance.initialize_code()
        instance.parameters.set_defaults
        N = 3
        t_begin = 0.0 | nbody_system.time
        t_end = 100.0 | nbody_system.time

        particles = p

        instance.particles.add_particles(particles)
        instance.commit_particles()

        sc = instance.stopping_conditions.collision_detection
        sc.enable()

        isCollision = False
        instance.evolve_model(t_end)
        isCollision = sc.is_set()

        instance.stop()
        self.assertTrue(isCollision, "no collision detected")



import numpy

from amuse.units import units, constants, nbody_system
from amuse.units.quantities import zero, AdaptingVectorQuantity, VectorQuantity
from amuse.support.exceptions import AmuseException
from amuse.datamodel import Particles
from amuse.datamodel import parameters
from amuse.ic.plummer import new_plummer_model

from amuse.test import amusetest
from amuse.couple import bridge


class TestCalculateFieldForParticles(amusetest.TestCase):

    def test1(self):
        particles = Particles(2)
        particles.mass = [1.0, 1.0] | nbody_system.mass
        particles.radius = [0.0001, 0.0001] | nbody_system.length
        particles.position = [[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]] | nbody_system.length
        particles.velocity = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]] | nbody_system.speed

        instance = bridge.CalculateFieldForParticles(particles=particles, gravity_constant=nbody_system.G)

        zero = 0.0 | nbody_system.length
        print(instance.get_gravity_at_point([zero], [1.0] | nbody_system.length, [zero], [zero]))
        fx, fy, fz = instance.get_gravity_at_point([zero], [1.0] | nbody_system.length, [zero], [zero])
        self.assertAlmostEqual(fx, [0.0] | nbody_system.acceleration, 6)
        self.assertAlmostEqual(fy, [0.0] | nbody_system.acceleration, 6)
        self.assertAlmostEqual(fz, [0.0] | nbody_system.acceleration, 6)

        for x in (0.25, 0.5, 0.75):
            x0 = x | nbody_system.length
            x1 = (2.0 - x) | nbody_system.length
            potential0 = instance.get_potential_at_point([zero], [x0], [zero], [zero])
            potential1 = instance.get_potential_at_point([zero], [x1], [zero], [zero])
            fx0, fy0, fz0 = instance.get_gravity_at_point([zero], [x0], [zero], [zero])
            fx1, fy1, fz1 = instance.get_gravity_at_point([zero], [x1], [zero], [zero])

            self.assertAlmostEqual(fy0[0], 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(fz0[0], 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(fy1[0], 0.0 | nbody_system.acceleration, 6)
            self.assertAlmostEqual(fz1[0], 0.0 | nbody_system.acceleration, 6)

            self.assertAlmostEqual(fx0, -1.0 * fx1, 5)
            fx = (-1.0 / (x0**2) + 1.0 / (x1**2)) * (1.0 | nbody_system.length ** 3 / nbody_system.time ** 2)
            self.assertAlmostEqual(fx, fx0[0], 5)
            self.assertAlmostEqual(potential0, potential1, 6)

    def test2(self):
        print("CalculateFieldForParticles, nbody units, no gravity_constant exceptions")
        stars = new_plummer_model(100)
        self.assertRaises(AmuseException, bridge.CalculateFieldForParticles, stars,
            expected_message="For generic units the gravity_constant must be specified")
        self.assertRaises(AmuseException, bridge.CalculateFieldForParticles, Particles(),
            expected_message="Particle data not yet available, so the gravity_constant must be specified")
        instance = bridge.CalculateFieldForParticles(particles=stars, gravity_constant=nbody_system.G)

    def test3(self):
        print("CalculateFieldForParticles get_potential_at_point, no softening")
        epsilon = 0 | units.m

        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)

        cluster = ExampleGravityCodeInterface()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)

        instance = bridge.CalculateFieldForParticles(particles=stars)
        instance.smoothing_length_squared = epsilon**2

        zeros = numpy.zeros(9) | units.parsec
        pos_range = numpy.linspace(-1.0, 1.0, 9) | units.parsec
        self.assertAlmostRelativeEqual(
            instance.get_potential_at_point(zeros, pos_range, zeros, zeros),
            cluster.get_potential_at_point(zeros, pos_range, zeros, zeros))
        for a_calculate_field, a_code in zip(
                instance.get_gravity_at_point(zeros, pos_range, zeros, zeros),
                cluster.get_gravity_at_point(zeros, pos_range, zeros, zeros)):
            self.assertAlmostRelativeEqual(a_calculate_field, a_code, 12)

    def test4(self):
        print("CalculateFieldForParticles get_potential_at_point, with softening")
        epsilon = 0.5 | units.parsec

        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)

        cluster = ExampleGravityCodeInterface()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)

        instance = bridge.CalculateFieldForParticles(particles=stars)
        instance.smoothing_length_squared = epsilon**2

        zeros = numpy.zeros(9) | units.parsec
        pos_range = numpy.linspace(-1.0, 1.0, 9) | units.parsec
        self.assertAlmostRelativeEqual(
            instance.get_potential_at_point(zeros, pos_range, zeros, zeros),
            cluster.get_potential_at_point(zeros, pos_range, zeros, zeros))
        for a_calculate_field, a_code in zip(
                instance.get_gravity_at_point(zeros, pos_range, zeros, zeros),
                cluster.get_gravity_at_point(zeros, pos_range, zeros, zeros)):
            self.assertAlmostRelativeEqual(a_calculate_field, a_code, 12)

    def test5(self):
        print("CalculateFieldForParticles get_potential_at_point, with individual softening")
        epsilon = 0.5 | units.parsec

        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)
        stars.radius = epsilon * numpy.random.uniform(low=0.4, high=3.0, size=len(stars))

        cluster = ExampleGravityCodeInterface(softening_mode="individual")
        cluster.particles.add_particles(stars)

        instance = bridge.CalculateFieldForParticles(particles=stars, softening_mode="individual")

        zeros = numpy.zeros(9) | units.parsec
        pos_range = numpy.linspace(-1.0, 1.0, 9) | units.parsec
        self.assertAlmostRelativeEqual(
            instance.get_potential_at_point(zeros, pos_range, zeros, zeros),
            cluster.get_potential_at_point(zeros, pos_range, zeros, zeros))
        for a_calculate_field, a_code in zip(
                instance.get_gravity_at_point(zeros, pos_range, zeros, zeros),
                cluster.get_gravity_at_point(zeros, pos_range, zeros, zeros)):
            self.assertAlmostRelativeEqual(a_calculate_field, a_code, 12)

    def test6(self):
        p = Particles(3)
        q = Particles(5)
        instance1 = bridge.CalculateFieldForParticles(gravity_constant="dummy")
        instance1.particles.add_particles(p)
        instance2 = bridge.CalculateFieldForParticles(gravity_constant="dummy")
        instance2.particles.add_particles(q)
        self.assertEqual(len(instance2.particles), 5)


class ExampleGravityCodeInterface(object):

    def __init__(self, softening_mode="shared"):
        self.particles = Particles()
        if softening_mode == "individual":
            self.softening_mode = "individual"
            self._softening_lengths_squared = self._softening_lengths_squared_individual
            self._softening_lengths = self._softening_lengths_individual
        else:
            self.softening_mode = "shared"
            self._softening_lengths_squared = self._softening_lengths_squared_shared
            self._softening_lengths = self._softening_lengths_shared
            epsilon_squared_parameter = parameters.ModuleMethodParameterDefinition(
                "get_epsilon_squared",
                "set_epsilon_squared",
                "epsilon_squared",
                "gravitational softening length squared",
                default_value=0.0 | nbody_system.length**2,
                must_set_before_get=False
            )
            self.parameters = parameters.new_parameters_instance_with_docs([epsilon_squared_parameter], self)
            self.epsilon_squared = 0.0 | nbody_system.length**2

    def _softening_lengths_squared_individual(self):
        return self.particles.radius**2

    def _softening_lengths_squared_shared(self):
        return self.epsilon_squared.as_vector_with_length(len(self.particles))

    def _softening_lengths_individual(self):
        return self.particles.radius

    def _softening_lengths_shared(self):
        return self.epsilon_squared.sqrt().as_vector_with_length(len(self.particles))

    def initialize_code(self):
        self.model_time = 0 | units.Myr

    def get_potential_at_point(self, eps, x, y, z):
        if isinstance(x, VectorQuantity):
            return -constants.G * (self.particles.mass.reshape((-1, 1)) /
                (self._softening_lengths_squared().reshape((-1, 1)) + eps.reshape((1, -1))**2 + (self.particles.x.reshape((-1, 1)) -
                x.reshape((1, -1)))**2 + (self.particles.y.reshape((-1, 1)) - y.reshape((1, -1)))**2 +
                (self.particles.z.reshape((-1, 1)) - z.reshape((1, -1)))**2).sqrt()).sum(axis=0)
        return -constants.G * (self.particles.mass / (self._softening_lengths_squared() + eps**2 + (self.particles.x - x)**2 +
            (self.particles.y - y)**2 + (self.particles.z - z)**2).sqrt()).sum()

    def get_gravity_at_point(self, eps, x, y, z):
        if isinstance(x, VectorQuantity):
            delta_x = x.reshape((1, -1)) - self.particles.x.reshape((-1, 1))
            delta_y = y.reshape((1, -1)) - self.particles.y.reshape((-1, 1))
            delta_z = z.reshape((1, -1)) - self.particles.z.reshape((-1, 1))
            factor = -constants.G * (self.particles.mass.reshape((-1, 1)) /
                (self._softening_lengths_squared().reshape((-1, 1)) + eps.reshape((1, -1))**2 +
                delta_x**2 + delta_y**2 + delta_z**2)**1.5)
            return (factor*delta_x).sum(axis=0), (factor*delta_y).sum(axis=0), (factor*delta_z).sum(axis=0)
        delta_x = self.particles.x - x
        delta_y = self.particles.y - y
        delta_z = self.particles.z - z
        factor = -constants.G * (self.particles.mass /
            (self._softening_lengths_squared() + eps**2 + delta_x**2 + delta_y**2 + delta_z**2)**1.5)
        return (factor*delta_x).sum(), (factor*delta_y).sum(), (factor*delta_z).sum()

    @property
    def potential_energy(self):
        if self.softening_mode == "individual":
            if len(self.particles) < 2:
                return zero * constants.G
            eps_vector = self.particles.radius**2
            sum_of_energies = zero
            for i in range(len(self.particles) - 1):
                dx = self.particles[i].x - self.particles[i+1:].x
                dy = self.particles[i].y - self.particles[i+1:].y
                dz = self.particles[i].z - self.particles[i+1:].z
                dr = ((dx * dx) + (dy * dy) + (dz * dz) + eps_vector[i] + eps_vector[i+1:]).sqrt()
                sum_of_energies -= (self.particles.mass[i] * self.particles.mass[i+1:] / dr).sum()
            return constants.G * sum_of_energies
        return self.particles.potential_energy(smoothing_length_squared=2*self.epsilon_squared)

    def before_set_parameter(self):
        pass

    def before_get_parameter(self):
        pass

    @property
    def kinetic_energy(self):
        return self.particles.kinetic_energy()

    def get_epsilon_squared(self):
        return self.epsilon_squared

    def set_epsilon_squared(self, epsilon_squared):
        self.epsilon_squared = epsilon_squared

    def commit_particles(self):
        self.set_accelerations()
        self.set_next_timestep()

    def set_accelerations(self):
        accelerations = self.get_gravity_at_point(self._softening_lengths(),
            self.particles.x, self.particles.y, self.particles.z)
        self.particles.ax = accelerations[0]
        self.particles.ay = accelerations[1]
        self.particles.az = accelerations[2]

    def set_next_timestep(self):
        self.next_timestep = min([0.01 * (self.particles.velocity /
            self.particles.acceleration).lengths_squared().amin().sqrt(), 1 | units.yr])

    def evolve_model(self, t_end):
        while self.model_time < t_end:
            dt = self.next_timestep
            self.particles.position += self.particles.velocity * dt + 0.5 * self.particles.acceleration * dt**2
            old_acceleration = self.particles.acceleration
            self.set_accelerations()
            self.particles.velocity += 0.5 * (old_acceleration + self.particles.acceleration) * dt
            self.model_time += dt
            self.set_next_timestep()


def system_from_particles(base_class, kwargs, particles, eps=None):
    interface = base_class(**kwargs)
    interface.initialize_code()
    if eps is not None:
        interface.parameters.epsilon_squared = eps**2
    interface.particles.add_particles(particles)
    interface.commit_particles()
    return interface


class TestBridge(amusetest.TestCase):

    def test1(self):
        print("Bridge potential energy with code's epsilon_squared as softening length")
        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        epsilon = 1.0e-2 | units.parsec
        test_class = ExampleGravityCodeInterface

        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)

        cluster = test_class()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)

        first_half = stars.select_array(lambda x: (x > 0 | units.m), ['x'])
        second_half = stars - first_half
        cluster1 = system_from_particles(test_class, dict(), first_half, epsilon)
        cluster2 = system_from_particles(test_class, dict(), second_half, epsilon)
        bridgesys = bridge.Bridge()
        bridgesys.add_system(cluster1, (cluster2,))
        bridgesys.add_system(cluster2, (cluster1,))

        self.assertAlmostRelativeEqual(cluster.potential_energy, bridgesys.potential_energy)
        self.assertAlmostRelativeEqual(cluster.kinetic_energy, bridgesys.kinetic_energy)

    def test2(self):
        print("Bridge potential energy with radius as softening length")
        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        epsilon = 0.1 | units.parsec
        test_class = ExampleGravityCodeInterface

        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)
        stars.radius = epsilon

        cluster = test_class()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)

        first_half = stars.select_array(lambda x: (x > 0 | units.m), ['x'])
        second_half = stars - first_half
        cluster1 = system_from_particles(test_class, dict(softening_mode="individual"), first_half)
        cluster2 = system_from_particles(test_class, dict(), second_half, epsilon)
        bridgesys = bridge.Bridge()
        bridgesys.add_system(cluster1, (cluster2,), radius_is_eps=True)
        bridgesys.add_system(cluster2, (cluster1,), radius_is_eps=True)

        self.assertAlmostRelativeEqual(cluster.potential_energy, bridgesys.potential_energy)
        self.assertAlmostRelativeEqual(cluster.kinetic_energy, bridgesys.kinetic_energy)

    def test3(self):
        print("Bridge potential energy with radius as softening length")
        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        epsilon = 0.1 | units.parsec
        test_class = ExampleGravityCodeInterface

        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)
        stars.radius = epsilon

        cluster = test_class()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)

        first_half = stars.select_array(lambda x: (x > 0 | units.m), ['x'])
        second_half = stars - first_half
        cluster1 = system_from_particles(test_class, dict(), first_half, epsilon)
        cluster2 = system_from_particles(test_class, dict(), second_half, epsilon)
        bridgesys = bridge.Bridge()
        bridgesys.add_system(cluster1, (cluster2,), radius_is_eps=True)
        bridgesys.add_system(cluster2, (cluster1,))

        self.assertAlmostRelativeEqual(cluster.potential_energy, bridgesys.potential_energy)
        self.assertAlmostRelativeEqual(cluster.kinetic_energy, bridgesys.kinetic_energy)

    def test4(self):
        print("Bridge evolve_model")
        convert = nbody_system.nbody_to_si(1.e5 | units.MSun, 1.0 | units.parsec)
        epsilon = 1.0e-2 | units.parsec
        test_class = ExampleGravityCodeInterface

        numpy.random.seed(12345)
        stars = new_plummer_model(100, convert_nbody=convert)

        cluster = test_class()
        cluster.initialize_code()
        cluster.parameters.epsilon_squared = epsilon**2
        cluster.particles.add_particles(stars)
        cluster.commit_particles()

        first_half = stars.select_array(lambda x: (x > 0 | units.m), ['x'])
        second_half = stars - first_half
        cluster1 = system_from_particles(test_class, dict(), first_half, epsilon)
        cluster2 = system_from_particles(test_class, dict(), second_half, epsilon)
        bridgesys = bridge.Bridge()
        bridgesys.add_system(cluster1, (cluster2,))
        bridgesys.add_system(cluster2, (cluster1,))

        self.assertAlmostRelativeEqual(
            cluster1.particles.get_intersecting_subset_in(cluster.particles).position,
            cluster1.particles.position)

# ~        old = cluster1.particles.position
        for i in range(2):
            one_timestep = cluster.next_timestep
            cluster.evolve_model(cluster.model_time + one_timestep)
            bridgesys.evolve_model(bridgesys.model_time + one_timestep, timestep=one_timestep)

# ~        print ((old - cluster1.particles.position)/old).lengths()
        self.assertAlmostRelativeEqual(
            cluster1.particles.get_intersecting_subset_in(cluster.particles).position,
            cluster1.particles.position)

        self.assertAlmostRelativeEqual(cluster.potential_energy, bridgesys.potential_energy)
        self.assertAlmostRelativeEqual(cluster.kinetic_energy, bridgesys.kinetic_energy)
